class PyTwitterError(Exception):
    """Base class for Twitter errors"""

    @property
    def message(self):
        """Returns the first argument used to construct this error."""
        return self.args[0]


class PythonTwitterDeprecationWarning(DeprecationWarning):
    """Base class for python-twitter deprecation warnings"""

    pass



"""
    tweet object

    Refer: https://developer.twitter.com/en/docs/twitter-api/data-dictionary/object-model/tweet
"""
import textwrap
from dataclasses import dataclass, field
from typing import Optional, List

from .base import BaseModel
from .stream import StreamRule


@dataclass
class TweetAttachments(BaseModel):
    poll_ids: Optional[List[str]] = field(default=None, repr=False, compare=False)
    media_keys: Optional[List[str]] = field(default=None, repr=False, compare=False)


@dataclass
class TweetContextAnnotationDomain(BaseModel):
    id: Optional[str] = field(default=None)
    name: Optional[str] = field(default=None)
    description: Optional[str] = field(default=None, repr=False)


@dataclass
class TweetContextAnnotationEntity(BaseModel):
    id: Optional[str] = field(default=None)
    name: Optional[str] = field(default=None)


@dataclass
class TweetContextAnnotation(BaseModel):
    """
    Refer https://developer.twitter.com/en/docs/twitter-api/annotations
    """

    domain: Optional[TweetContextAnnotationDomain] = field(
        default=None, repr=False, compare=False
    )
    entity: Optional[TweetContextAnnotationEntity] = field(
        default=None, repr=False, compare=False
    )


@dataclass
class TweetEntitiesAnnotation(BaseModel):
    start: Optional[int] = field(default=None, repr=False)
    end: Optional[int] = field(default=None, repr=False)
    probability: Optional[float] = field(default=None, repr=False)
    type: Optional[str] = field(default=None)
    normalized_text: Optional[str] = field(default=None)


@dataclass
class TweetEntitiesHashtag(BaseModel):
    start: Optional[int] = field(default=None, repr=False)
    end: Optional[int] = field(default=None, repr=False)
    tag: Optional[str] = field(default=None)


@dataclass
class TweetEntitiesCashtag(TweetEntitiesHashtag):
    ...


@dataclass
class TweetEntitiesMention(BaseModel):
    start: Optional[int] = field(default=None, repr=False)
    end: Optional[int] = field(default=None, repr=False)
    username: Optional[str] = field(default=None)


@dataclass
class TweetEntitiesUrl(BaseModel):
    start: Optional[int] = field(default=None, repr=False)
    end: Optional[int] = field(default=None, repr=False)
    url: Optional[str] = field(default=None)
    expanded_url: Optional[str] = field(default=None, repr=False)
    display_url: Optional[str] = field(default=None, repr=False)
    status: Optional[str] = field(default=None, repr=False)
    title: Optional[str] = field(default=None, repr=False)
    description: Optional[str] = field(default=None, repr=False)
    unwound_url: Optional[str] = field(default=None, repr=False)


@dataclass
class TweetEntities(BaseModel):
    annotations: Optional[List[TweetEntitiesAnnotation]] = field(
        default=None, repr=False
    )
    cashtags: Optional[List[TweetEntitiesCashtag]] = field(default=None, repr=False)
    hashtags: Optional[List[TweetEntitiesHashtag]] = field(default=None, repr=False)
    mentions: Optional[List[TweetEntitiesMention]] = field(default=None, repr=False)
    urls: Optional[List[TweetEntitiesUrl]] = field(default=None, repr=False)


@dataclass
class TweetGeoCoordinates(BaseModel):
    type: Optional[str] = field(default=None)
    coordinates: Optional[List[float]] = field(default=None, repr=False)


@dataclass
class TweetGeo(BaseModel):
    coordinates: Optional[TweetGeoCoordinates] = field(default=None, repr=False)
    place_id: Optional[str] = field(default=None)


@dataclass
class TweetNonPublicMetrics(BaseModel):
    impression_count: Optional[int] = field(default=None)
    url_link_clicks: Optional[int] = field(default=None, repr=False)
    user_profile_clicks: Optional[int] = field(default=None, repr=False)


@dataclass
class TweetOrganicMetrics(TweetNonPublicMetrics):
    like_count: Optional[int] = field(default=None, repr=False)
    reply_count: Optional[int] = field(default=None, repr=False)
    retweet_count: Optional[int] = field(default=None, repr=False)


@dataclass
class TweetPromotedMetrics(TweetOrganicMetrics):
    ...


@dataclass
class TweetPublicMetrics(BaseModel):
    retweet_count: Optional[int] = field(default=None, repr=False)
    quote_count: Optional[int] = field(default=None, repr=False)
    reply_count: Optional[int] = field(default=None, repr=False)
    like_count: Optional[int] = field(default=None)


@dataclass
class TweetWithheld(BaseModel):
    """
    refer: https://help.twitter.com/en/rules-and-policies/tweet-withheld-by-country
    """

    copyright: Optional[bool] = field(default=None, repr=False)
    country_codes: Optional[List[str]] = field(default=None, repr=False, compare=False)


@dataclass
class TweetReferencedTweet(BaseModel):
    type: Optional[str] = field(default=None, repr=False, compare=False)
    id: Optional[str] = field(default=None, repr=False, compare=False)


@dataclass
class Tweet(BaseModel):
    """
    A class representing the tweet object.
    """

    id: Optional[str] = field(default=None)
    text: Optional[str] = field(default=None)
    attachments: Optional[TweetAttachments] = field(
        default=None, repr=False, compare=False
    )
    author_id: Optional[str] = field(default=None, repr=False, compare=False)
    context_annotations: Optional[List[TweetContextAnnotation]] = field(
        default=None, repr=False, compare=False
    )
    conversation_id: Optional[str] = field(default=None, repr=False, compare=False)
    created_at: Optional[str] = field(default=None, repr=False, compare=False)
    entities: Optional[TweetEntities] = field(default=None, repr=False, compare=False)
    geo: Optional[TweetGeo] = field(default=None, repr=False, compare=False)
    in_reply_to_user_id: Optional[str] = field(default=None, repr=False, compare=False)
    lang: Optional[str] = field(default=None, repr=False, compare=False)
    non_public_metrics: Optional[TweetNonPublicMetrics] = field(
        default=None, repr=False, compare=False
    )
    organic_metrics: Optional[TweetNonPublicMetrics] = field(
        default=None, repr=False, compare=False
    )
    possibly_sensitive: Optional[bool] = field(default=None, repr=False, compare=False)
    promoted_metrics: Optional[TweetPromotedMetrics] = field(
        default=None, repr=False, compare=False
    )
    public_metrics: Optional[TweetPublicMetrics] = field(
        default=None, repr=False, compare=False
    )
    referenced_tweets: Optional[List[TweetReferencedTweet]] = field(
        default=None, repr=False, compare=False
    )
    reply_settings: Optional[str] = field(default=None, repr=False, compare=False)
    source: Optional[str] = field(default=None, repr=False, compare=False)
    withheld: Optional[TweetWithheld] = field(default=None, repr=False, compare=False)

    # Note: this field only for stream tweet
    matching_rules: Optional[List[StreamRule]] = field(
        default=None, repr=False, compare=False
    )

    def __repr__(self):
        text = self.text
        if text:
            # Make text shorter to display.
            text = textwrap.shorten(self.text, width=50, placeholder="...")
        return f"Tweet(id={self.id}, text={text})"



"""
    Twitter API Rate Limit
"""
import logging
import re
from collections import defaultdict
from dataclasses import dataclass
from typing import Optional, Pattern
from urllib.parse import urlparse

from pytwitter.error import PyTwitterError
from pytwitter.utils.convertors import conv_type

logger = logging.getLogger(__name__)


@dataclass
class RateLimitData:
    limit: int = 15
    remaining: int = 15
    reset: int = 0


@dataclass
class Endpoint:
    resource: str
    regex: Optional[Pattern[str]] = None

    # For different auth type and endpoint
    LIMIT_APP_GET: int = 15
    LIMIT_USER_GET: int = 15
    LIMIT_APP_POST: int = 0
    LIMIT_USER_POST: int = 0
    LIMIT_USER_PUT: int = 0
    LIMIT_USER_DELETE: int = 0

    def get_limit(self, auth_type, method="GET"):
        return getattr(self, f"LIMIT_{auth_type.upper()}_{method}", 0)


TWEETS_BY_ID = Endpoint(
    resource="/tweets",
    regex=re.compile(r"/tweets"),
    LIMIT_APP_GET=300,
    LIMIT_USER_GET=900,
    LIMIT_USER_POST=200,
)
TWEET_BY_ID = Endpoint(
    resource="/tweets/:id",
    regex=re.compile(r"/tweets/\d+"),
    LIMIT_APP_GET=300,
    LIMIT_USER_GET=900,
    LIMIT_USER_DELETE=50,
)
USER_TIMELINE = Endpoint(
    resource="/users/:id/tweets",
    regex=re.compile(r"/users/\d+/tweets"),
    LIMIT_APP_GET=1500,
    LIMIT_USER_GET=900,
)
USER_MENTIONS = Endpoint(
    resource="/users/:id/mentions",
    regex=re.compile(r"/users/\d+/mentions"),
    LIMIT_APP_GET=450,
    LIMIT_USER_GET=180,
)
TWEET_SEARCH_RECENT = Endpoint(
    resource="/tweets/search/recent",
    regex=re.compile(r"/tweets/search/recent"),
    LIMIT_APP_GET=450,
    LIMIT_USER_GET=180,
)
TWEET_SEARCH_ALL = Endpoint(
    resource="/tweets/search/all",
    regex=re.compile(r"/tweets/search/all"),
    LIMIT_APP_GET=300,
)
TWEET_COUNTS = Endpoint(
    resource="/tweets/counts",
    regex=re.compile(r"/tweets/counts/\w+"),
    LIMIT_APP_GET=300,
)
TWEET_RETWEET_USER = Endpoint(
    resource="/tweets/:id/retweeted_by",
    regex=re.compile(r"/tweets/\d+/retweeted_by"),
    LIMIT_APP_GET=75,
    LIMIT_USER_GET=75,
)
USER_TWEET_RETWEET = Endpoint(
    resource="/users/:id/retweets",
    regex=re.compile(r"/users/\d+/retweets"),
    LIMIT_USER_POST=50,
)
USER_TWEET_RETWEET_REMOVE = Endpoint(
    resource="/users/:id/retweets/:tweet_id",
    regex=re.compile(r"/users/\d+/retweets/\d+"),
    LIMIT_USER_POST=50,
)
TWEET_LIKING_USER = Endpoint(
    resource="/tweets/:id/liking_users",
    regex=re.compile(r"/tweets/\d+/liking_users"),
    LIMIT_APP_GET=75,
    LIMIT_USER_GET=75,
)
USER_LIKED_TWEET = Endpoint(
    resource="/users/:id/liked_tweets",
    regex=re.compile(r"/users/\d+/liked_tweets"),
    LIMIT_APP_GET=75,
    LIMIT_USER_GET=75,
)
USER_TWEET_LIKE = Endpoint(
    resource="/users/:id/likes",
    regex=re.compile(r"/users/\d+/likes"),
    LIMIT_USER_POST=50,
)
USER_TWEET_LIKE_REMOVE = Endpoint(
    resource="/users/:id/likes/:tweet_id",
    regex=re.compile(r"/users/\d+/likes/\d+"),
    LIMIT_USER_POST=50,
)
TWEET_HIDDEN = Endpoint(
    resource="/tweets/:id/hidden",
    regex=re.compile(r"/tweets/\d+/hidden"),
    LIMIT_USER_PUT=50,
)

USER_BY_ID = Endpoint(
    resource="/users/:id",
    regex=re.compile(r"/users/\d+"),
    LIMIT_APP_GET=300,
    LIMIT_USER_GET=900,
)
USERS_BY_ID = Endpoint(
    resource="/users",
    regex=re.compile(r"/users"),
    LIMIT_APP_GET=300,
    LIMIT_USER_GET=900,
)
USER_BY_USERNAME = Endpoint(
    resource="/users/by/:username",
    regex=re.compile(r"/users/by/\w+"),
    LIMIT_APP_GET=300,
    LIMIT_USER_GET=900,
)
USERS_BY_USERNAME = Endpoint(
    resource="/users/by",
    regex=re.compile(r"/users/by"),
    LIMIT_APP_GET=300,
    LIMIT_USER_GET=900,
)
USER_FOLLOWING = Endpoint(
    resource="/users/:id/following",
    regex=re.compile(r"/users/\d+/following"),
    LIMIT_APP_GET=15,
    LIMIT_USER_GET=15,
    LIMIT_USER_POST=15,
)
USER_REMOVE_FOLLOWING = Endpoint(
    resource="/users/:id/following/:target_user_id",
    regex=re.compile(r"/users/\d+/following/\d+"),
    LIMIT_USER_DELETE=50,
)
USER_FOLLOWER = Endpoint(
    resource="/users/:id/followers",
    regex=re.compile(r"/users/\d+/followers"),
    LIMIT_APP_GET=15,
    LIMIT_USER_GET=15,
)
USER_BLOCKING = Endpoint(
    resource="/users/:id/blocking",
    regex=re.compile(r"/users/\d+/blocking"),
    LIMIT_USER_GET=15,
    LIMIT_USER_POST=50,
)
USER_REMOVE_BLOCKING = Endpoint(
    resource="/users/:id/blocking/:target_user_id",
    regex=re.compile(r"/users/\d+/blocking/\d+"),
    LIMIT_USER_DELETE=50,
)
USER_MUTING = Endpoint(
    resource="/users/:id/muting",
    regex=re.compile(r"/users/\d+/muting"),
    LIMIT_USER_GET=15,
    LIMIT_USER_POST=50,
)
USER_REMOVE_MUTING = Endpoint(
    resource="/users/:id/muting/:target_user_id",
    regex=re.compile(r"/users/\d+/muting/\d+"),
    LIMIT_USER_POST=50,
)

SPACE_BY_ID = Endpoint(
    resource="/spaces/:id",
    regex=re.compile(r"/spaces/\w+"),
    LIMIT_APP_GET=300,
)
SPACES_BY_IDS = Endpoint(
    resource="/spaces",
    regex=re.compile(r"/spaces"),
    LIMIT_APP_GET=300,
)
SPACES_BY_CREATORS = Endpoint(
    resource="/spaces/by/creator_ids",
    regex=re.compile(r"/spaces/by/creator_ids"),
    LIMIT_APP_GET=300,
)
SPACES_SEARCH = Endpoint(
    resource="/spaces/search",
    regex=re.compile("/spaces/search"),
    LIMIT_APP_GET=300,
)

LIST_BY_ID = Endpoint(
    resource="/lists/:id",
    regex=re.compile(r"/lists/\d+"),
    LIMIT_APP_GET=75,
    LIMIT_USER_GET=75,
)
LISTS_BY_USER = Endpoint(
    resource="/users/:id/owned_lists",
    regex=re.compile(r"/users/\d+/owned_lists"),
    LIMIT_APP_GET=15,
    LIMIT_USER_GET=15,
)
LISTS_MANAGE = Endpoint(
    resource="/lists",
    regex=re.compile(r"/lists"),
    LIMIT_USER_POST=300,
)
LISTS_MANAGE_BY_ID = Endpoint(
    resource="/lists/:id",
    regex=re.compile(r"/lists/\w+"),
    LIMIT_USER_PUT=300,
    LIMIT_USER_DELETE=300,
)
LIST_TWEETS = Endpoint(
    resource="/lists/:id/tweets",
    regex=re.compile(r"/lists/\d+/tweets"),
    LIMIT_APP_GET=900,
    LIMIT_USER_GET=900,
)
LIST_MEMBERS = Endpoint(
    resource="/lists/:id/members",
    regex=re.compile(r"/lists/\d+/members"),
    LIMIT_APP_GET=900,
    LIMIT_USER_GET=900,
)
LISTS_MEMBERSHIPS_BY_USER = Endpoint(
    resource="/users/:id/list_memberships",
    regex=re.compile(r"/users/\d+/list_memberships"),
    LIMIT_APP_GET=75,
    LIMIT_USER_GET=75,
)
LISTS_MANAGE_MEMBERS = Endpoint(
    resource="/lists/members",
    regex=re.compile(r"/lists/members"),
    LIMIT_USER_POST=300,
)
LISTS_MANAGE_MEMBERS_BY_ID = Endpoint(
    resource="/lists/:id/members/:user_id",
    regex=re.compile(r"/lists/\w+/members/\d+"),
    LIMIT_USER_DELETE=300,
)
LISTS_MANAGE_FOLLOWERS = Endpoint(
    resource="/lists/followers",
    regex=re.compile(r"/lists/followers"),
    LIMIT_USER_GET=180,
    LIMIT_APP_GET=180,
)
USER_FOLLOW_LISTS = Endpoint(
    resource="/users/:id/followed_lists",
    regex=re.compile(r"/users/\d+/followed_lists"),
    LIMIT_USER_POST=50,
    LIMIT_USER_GET=15,
    LIMIT_APP_GET=15,
)
USER_FOLLOW_LISTS_BY_ID = Endpoint(
    resource="/users/:id/followed_lists/:list_id",
    regex=re.compile(r"/users/\d+/followed_lists/\w+"),
    LIMIT_USER_DELETE=50,
)
USER_PINNED_LISTS = Endpoint(
    resource="/users/:id/pinned_lists",
    regex=re.compile(r"/users/\d+/pinned_lists"),
    LIMIT_USER_GET=15,
    LIMIT_USER_POST=50,
)
USER_PINNED_LISTS_BY_ID = Endpoint(
    resource="/users/:id/pinned_lists/:list_id",
    regex=re.compile(r"/users/\d+/pinned_lists/\w+"),
    LIMIT_USER_DELETE=50,
)

COMPLIANCE_JOB_BY_ID = Endpoint(
    resource="/compliance/jobs/:job_id",
    regex=re.compile(r"/compliance/jobs/\d+"),
    LIMIT_APP_GET=150,
)
COMPLIANCE_JOBS = Endpoint(
    resource="/compliance/jobs",
    regex=re.compile(r"/compliance/jobs"),
    LIMIT_APP_GET=150,
    LIMIT_APP_POST=150,
)

PATH_VAR_ENDPOINTS = [
    TWEET_BY_ID,
    TWEETS_BY_ID,
    USER_TIMELINE,
    USER_MENTIONS,
    TWEET_SEARCH_RECENT,
    TWEET_SEARCH_ALL,
    TWEET_COUNTS,
    TWEET_RETWEET_USER,
    USER_TWEET_RETWEET,
    USER_TWEET_RETWEET_REMOVE,
    TWEET_LIKING_USER,
    USER_LIKED_TWEET,
    USER_TWEET_LIKE,
    USER_TWEET_LIKE_REMOVE,
    TWEET_HIDDEN,
    USER_BY_ID,
    USERS_BY_ID,
    USER_BY_USERNAME,
    USERS_BY_USERNAME,
    USER_FOLLOWING,
    USER_REMOVE_FOLLOWING,
    USER_FOLLOWER,
    USER_BLOCKING,
    USER_REMOVE_BLOCKING,
    USER_MUTING,
    USER_REMOVE_MUTING,
    SPACE_BY_ID,
    SPACES_BY_IDS,
    SPACES_BY_CREATORS,
    SPACES_SEARCH,
    LIST_BY_ID,
    LISTS_BY_USER,
    LISTS_MANAGE,
    LISTS_MANAGE_BY_ID,
    LIST_TWEETS,
    LIST_MEMBERS,
    LISTS_MEMBERSHIPS_BY_USER,
    LISTS_MANAGE_MEMBERS,
    LISTS_MANAGE_MEMBERS_BY_ID,
    LISTS_MANAGE_FOLLOWERS,
    USER_FOLLOW_LISTS,
    USER_FOLLOW_LISTS_BY_ID,
    USER_PINNED_LISTS,
    USER_PINNED_LISTS_BY_ID,
    COMPLIANCE_JOB_BY_ID,
    COMPLIANCE_JOBS,
]


class RateLimit:
    """
    API rate limit.
    Refer: https://developer.twitter.com/en/docs/twitter-api/rate-limits
    """

    def __init__(self, auth_type="app"):
        """
        Stored rate limit data. like:
        ``` json
        {
            "/users/:id": RateLimitData(limit=300, remaining=299, reset=1612522034),
            "/users/by/:username": RateLimitData(limit=300, remaining=289, reset=1612522029),
        }
        ```
        :param auth_type: app auth or user auth
        """
        if auth_type.lower() not in ("user", "app"):
            raise PyTwitterError(f"Not support for auth type {auth_type}")
        self.auth_type = auth_type
        self.mapping = defaultdict(dict)

    @staticmethod
    def url_to_endpoint(url) -> Endpoint:
        resource = urlparse(url).path.replace("/2", "", 1)  # only replace api version
        for endpoint in PATH_VAR_ENDPOINTS:
            if re.fullmatch(endpoint.regex, resource):
                return endpoint
        return Endpoint(resource=resource)

    def set_limit(self, url, headers, method="GET") -> RateLimitData:
        """
        Twitter API rate limit data stored at requests headers. Like:

        ``` json
        {
            "x-rate-limit-limit": "300",
            "x-rate-limit-remaining": "299",
            "x-rate-limit-reset": "1612519043",
        }
        ```

        :param url: api query url.
        :param headers: api response headers.
        :param method: request method
        :return:
        """
        endpoint = self.url_to_endpoint(url=url)
        data = {
            "limit": conv_type("limit", int, headers.get("x-rate-limit-limit", 0)),
            "remaining": conv_type(
                "remaining", int, headers.get("x-rate-limit-remaining", 0)
            ),
            "reset": conv_type("reset", int, headers.get("x-rate-limit-reset", 0)),
        }
        self.mapping[endpoint.resource][method.upper()] = RateLimitData(**data)

        return self.get_limit(url=url, method=method)

    def get_limit(self, url, method="GET") -> RateLimitData:
        endpoint = self.url_to_endpoint(url=url)
        if method not in self.mapping.get(endpoint.resource, {}):
            limit = endpoint.get_limit(auth_type=self.auth_type, method=method)
            return RateLimitData(limit=limit, remaining=limit)
        return self.mapping[endpoint.resource][method.upper()]



"""
    Place object

    Refer: https://developer.twitter.com/en/docs/twitter-api/data-dictionary/object-model/place
"""

from dataclasses import dataclass, field, make_dataclass
from typing import Dict, List, Optional, Union

from .base import BaseModel


@dataclass
class PlaceGeoProperties(BaseModel):
    name: Optional[str] = field(default=None, repr=False)


@dataclass
class PlaceGeo(BaseModel):
    type: Optional[str] = field(default=None)
    bbox: Optional[List[float]] = field(default=None, repr=False)
    properties: Optional[Union[Dict, PlaceGeoProperties]] = field(
        default=None, repr=False
    )

    def __post_init__(self):
        if self.properties is not None and isinstance(self.properties, dict):
            new_properties = make_dataclass(
                "PlaceGeoProperties",
                [
                    (key, Optional[type(value)], field(default=None, repr=False))
                    for key, value in self.properties.items()
                ],
                bases=(PlaceGeoProperties,),
            )
            self.properties = new_properties.new_from_json_dict(self.properties)


@dataclass
class Place(BaseModel):
    """
    A class representing place object.
    """

    id: Optional[str] = field(default=None)
    full_name: Optional[str] = field(default=None)
    country: Optional[str] = field(default=None, repr=False)
    country_code: Optional[str] = field(default=None, repr=False)
    geo: Optional[PlaceGeo] = field(default=None, repr=False)
    name: Optional[str] = field(default=None, repr=False)
    place_type: Optional[str] = field(default=None, repr=False)



"""
    Api for streaming.
"""
import base64
import json
import logging
import time
from typing import Dict, List, Optional, Set, Tuple, Union

import requests
import pytwitter.models as md
from pytwitter.error import PyTwitterError
from pytwitter.utils.validators import enf_comma_separated
from requests.models import Response
from authlib.integrations.requests_client import OAuth2Auth

logger = logging.getLogger(__name__)


class StreamApi:
    BASE_URL = "https://api.twitter.com/2"

    def __init__(
        self,
        bearer_token=None,
        consumer_key=None,
        consumer_secret=None,
        proxies=None,
        max_retries=3,
        timeout=None,
        chunk_size=1024,
    ):
        self.running = False
        self._auth = None
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.proxies = proxies
        self.session = requests.Session()
        self.max_retries = max_retries
        self.timeout = timeout
        self.chunk_size = chunk_size

        if bearer_token:
            self._auth = OAuth2Auth(
                token={"access_token": bearer_token, "token_type": "Bearer"}
            )
        elif all([self.consumer_key, self.consumer_secret]):
            resp = self.generate_bearer_token(
                consumer_key=consumer_key, consumer_secret=consumer_secret
            )
            self._auth = OAuth2Auth(
                token={"access_token": resp["access_token"], "token_type": "Bearer"}
            )
        else:
            raise PyTwitterError("Need oauth")

    @staticmethod
    def generate_bearer_token(consumer_key: str, consumer_secret: str) -> dict:
        """
        :param consumer_key: Your app consumer key
        :param consumer_secret: Your app consumer secret
        :return: token data
        """
        bearer_token = base64.b64encode(f"{consumer_key}:{consumer_secret}".encode())
        headers = {
            "Authorization": f"Basic {bearer_token.decode()}",
            "Content-Type": "application/x-www-form-urlencoded;charset=UTF-8",
        }
        resp = requests.post(
            url="https://api.twitter.com/oauth2/token",
            data={"grant_type": "client_credentials"},
            headers=headers,
        )

        data = resp.json()
        if "errors" in data:
            raise PyTwitterError(data["errors"])
        return data

    def _connect(self, url, params=None, return_json=False):
        """
        :param url:
        :param params:
        :param return_json:
        :return:
        """
        # make sure only one running connect
        self.running = True
        retries = 0
        http_error_wait = 5
        http_error_wait_max = 320

        try:
            while self.running and retries < self.max_retries:
                with self.session.get(
                    url=url,
                    params=params,
                    auth=self._auth,
                    proxies=self.proxies,
                    timeout=self.timeout,
                    stream=True,
                ) as resp:
                    logger.debug(resp.headers)
                    if resp.status_code == 200:
                        for line in resp.iter_lines(chunk_size=self.chunk_size):
                            if line:
                                self.on_data(raw_data=line, return_json=return_json)
                            else:
                                self.on_keep_alive()
                            if not self.running:
                                break

                        if resp.raw.closed:
                            self.on_closed(resp)
                    else:
                        self.on_request_error(resp)
                        retries += 1
                        time.sleep(http_error_wait)

                        http_error_wait *= 2
                        if http_error_wait > http_error_wait_max:
                            break
        except Exception as exc:
            logger.exception(f"Exception in request, exc: {exc}")
        finally:
            self.session.close()
            self.running = False

    def disconnect(self):
        self.running = False

    def on_data(self, raw_data, return_json=False):
        """
        :param raw_data: Response data by twitter api.
        :param return_json:
        :return:
        """
        data = json.loads(raw_data)
        if not return_json:
            data = data.get("data")
            data = md.Tweet.new_from_json_dict(data=data)

        return self.on_tweet(tweet=data)

    def on_tweet(self, tweet):
        """
        :param tweet: Tweet obj or json data.
        :return:
        """
        logger.debug(f"Received tweet: {tweet}")

    def on_keep_alive(self):
        """
        Refer: https://developer.twitter.com/en/docs/twitter-api/tweets/sampled-stream/integrate/handling-disconnections
        :return:
        """
        logger.debug("Received keep alive signal")

    def on_request_error(self, resp):
        logger.debug(f"Received error status code: {resp.status_code}")

    def on_closed(self, resp):
        logger.debug("Received closed response")

    def sample_stream(
        self,
        *,
        backfill_minutes: Optional[int] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        media_fields: Optional[Union[str, List, Tuple]] = None,
        place_fields: Optional[Union[str, List, Tuple]] = None,
        poll_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ):
        """
        Streams about 1% of all Tweets in real-time.

        :param backfill_minutes: Minutes for disconnection with reconnected stream.
            Accepted value is 1 to 5.
        :param tweet_fields: Fields for the tweet object.
        :param expansions: Fields for the expansions.
        :param user_fields: Fields for the user object.
        :param media_fields: Fields for the media object.
        :param place_fields: Fields for the place object.
        :param poll_fields: Fields for the poll object.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        """

        if self.running:
            raise PyTwitterError("Stream is running")

        args = {
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "media.fields": enf_comma_separated(
                name="media_fields", value=media_fields
            ),
            "place.fields": enf_comma_separated(
                name="place_fields", value=place_fields
            ),
            "poll.fields": enf_comma_separated(name="poll_fields", value=poll_fields),
        }
        if backfill_minutes is not None:
            args["backfill_minutes"] = backfill_minutes

        # connect the stream
        self._connect(
            url=f"{self.BASE_URL}/tweets/sample/stream",
            params=args,
            return_json=return_json,
        )

    def search_stream(
        self,
        *,
        backfill_minutes: Optional[int] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        media_fields: Optional[Union[str, List, Tuple]] = None,
        place_fields: Optional[Union[str, List, Tuple]] = None,
        poll_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ):

        """
        Streams Tweets in real-time based on a specific set of filter rules.

        :param backfill_minutes: Minutes for disconnection with reconnected stream.
            Accepted value is 1 to 5.
        :param tweet_fields: Fields for the tweet object.
        :param expansions: Fields for the expansions.
        :param user_fields: Fields for the user object.
        :param media_fields: Fields for the media object.
        :param place_fields: Fields for the place object.
        :param poll_fields: Fields for the poll object.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        """

        if self.running:
            raise PyTwitterError("Stream is running")

        args = {
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "media.fields": enf_comma_separated(
                name="media_fields", value=media_fields
            ),
            "place.fields": enf_comma_separated(
                name="place_fields", value=place_fields
            ),
            "poll.fields": enf_comma_separated(name="poll_fields", value=poll_fields),
        }
        if backfill_minutes is not None:
            args["backfill_minutes"] = backfill_minutes

        # connect the stream
        self._connect(
            url=f"{self.BASE_URL}/tweets/search/stream",
            params=args,
            return_json=return_json,
        )

    def _request(self, url, verb="GET", params=None, json_data=None):
        """
        :param url: Url for twitter api
        :param verb: HTTP Method, like GET,POST.
        :param params: The url params to send in the body of the request.
        :param json_data: The json data to send in the body of the request.
        :return: A json object
        """

        resp = self.session.request(
            url=url,
            method=verb,
            params=params,
            auth=self._auth,
            json=json_data,
            timeout=self.timeout,
            proxies=self.proxies,
        )

        return resp

    @staticmethod
    def _parse_response(resp: Response) -> dict:
        """
        :param resp: Response
        :return: json data
        """
        try:
            data = resp.json()
        except ValueError:
            raise PyTwitterError(f"Unknown error: {resp.content}")

        if not resp.ok:
            raise PyTwitterError(data)

        return data

    def get_rules(
        self, ids: Optional[Union[str, List, Tuple]] = None, return_json=False
    ):
        """
        Return a list of rules currently active on the streaming endpoint, either as a list or individually.

        :param ids: IDs for rule. If omitted, all rules are returned.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return: Response object or json data
        """

        args = {"ids": enf_comma_separated(name="ids", value=ids)}

        resp = self._request(
            url=f"{self.BASE_URL}/tweets/search/stream/rules",
            params=args,
        )
        resp_json = self._parse_response(resp=resp)

        if return_json:
            return resp_json
        else:
            return md.Response(
                data=[
                    md.StreamRule.new_from_json_dict(item)
                    for item in resp_json.get("data", [])
                ],
                meta=md.Meta.new_from_json_dict(resp_json.get("meta")),
            )

    def manage_rules(
        self, rules: Optional[Dict[str, List]], dry_run=False, return_json=False
    ):
        """
        Add or delete rules to your stream.

        :param rules: Json body for your rules.
            See more detail: https://developer.twitter.com/en/docs/twitter-api/tweets/filtered-stream/api-reference/post-tweets-search-stream-rules
        :param dry_run: Set to true can test the syntax of your rules without submitting it.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return: Response object or json data
        """

        resp = self._request(
            url=f"{self.BASE_URL}/tweets/search/stream/rules",
            verb="POST",
            params={"dry_run": dry_run},
            json_data=rules,
        )

        resp_json = self._parse_response(resp=resp)

        if return_json:
            return resp_json
        else:
            errors = resp_json.get("errors")
            return md.Response(
                data=[
                    md.StreamRule.new_from_json_dict(item)
                    for item in resp_json.get("data", [])
                ],
                meta=md.Meta.new_from_json_dict(resp_json.get("meta")),
                errors=[md.Error.new_from_json_dict(err) for err in errors]
                if errors
                else None,
            )



"""
    stream obj.
"""

from dataclasses import dataclass, field
from typing import Optional

from .base import BaseModel


@dataclass
class StreamRule(BaseModel):
    id: Optional[str] = field(default=None)
    value: Optional[str] = field(default=None)
    tag: Optional[str] = field(default=None, repr=False)



"""
    Api Impl
"""
import base64
import logging
import os
import re
import time
from typing import List, Optional, Tuple, Union

import requests
from requests.models import Response
from authlib.integrations.requests_client import (
    OAuth2Session,
    OAuth2Auth,
    OAuth1Auth,
    OAuth1Session,
)

import pytwitter.models as md
from pytwitter.error import PyTwitterError
from pytwitter.rate_limit import RateLimit
from pytwitter.utils.validators import enf_comma_separated

logger = logging.getLogger(__name__)


class Api:
    BASE_URL_V2 = "https://api.twitter.com/2"
    BASE_REQUEST_TOKEN_URL = "https://api.twitter.com/oauth/request_token"
    BASE_AUTHORIZE_URL = "https://api.twitter.com/oauth/authorize"
    BASE_ACCESS_TOKEN_URL = "https://api.twitter.com/oauth/access_token"
    DEFAULT_CALLBACK_URI = "https://localhost/"
    BASE_OAUTH2_AUTHORIZE_URL = "https://twitter.com/i/oauth2/authorize"
    BASE_OAUTH2_ACCESS_TOKEN_URL = "https://api.twitter.com/2/oauth2/token"
    DEFAULT_SCOPES = ["users.read", "tweet.read"]

    def __init__(
        self,
        bearer_token: Optional[str] = None,
        consumer_key: Optional[str] = None,
        consumer_secret: Optional[str] = None,
        access_token: Optional[str] = None,
        access_secret: Optional[str] = None,
        client_id: Optional[str] = None,
        application_only_auth: bool = False,
        oauth_flow: bool = False,  # provide access with authorize
        sleep_on_rate_limit: bool = False,
        timeout: Optional[int] = None,
        proxies: Optional[dict] = None,
        callback_uri: Optional[str] = None,
        scopes: Optional[List[str]] = None,
    ) -> None:
        """
        Initial the Api instance.

        :param bearer_token: OAuth2.0 Token for app.
        :param consumer_key: Consumer key for app.
        :param consumer_secret: Consumer secret for app.
        :param access_token: Access token for the authenticated user.
        :param access_secret: Access token secret for the authenticated user.
        :param client_id: Client ID for app with OAuth2.0
        :param application_only_auth: If set this, with auto exchange app bearer token with consumer credentials.
        :param oauth_flow: If set this, You need generate access token with user by OAuth1.1 or OAuth2.0
        :param sleep_on_rate_limit: If token reach the limit, will sleep.
        :param timeout: Timeout for requests.
        :param proxies: Proxies for requests.
        :param callback_uri: Your callback URL. This value must correspond to one of the Callback URLs defined in your App settings.
        :param scopes: Scopes allow you to set granular access for your App so that your App only has the permissions that it needs.
        """
        self.session = requests.Session()
        self._auth = None
        self._oauth_session = None
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.client_id = client_id
        self.timeout = timeout
        self.proxies = proxies
        self.rate_limit = RateLimit()
        self.sleep_on_rate_limit = sleep_on_rate_limit
        self.auth_user_id = None  # Note: use this keep uid for auth user
        self.callback_uri = (
            callback_uri if callback_uri is not None else self.DEFAULT_CALLBACK_URI
        )
        self.scopes = scopes if scopes is not None else self.DEFAULT_SCOPES

        # just use bearer token
        if bearer_token:
            self._auth = OAuth2Auth(
                token={"access_token": bearer_token, "token_type": "Bearer"}
            )
        # use app auth
        elif consumer_key and consumer_secret and application_only_auth:
            resp = self.generate_bearer_token(
                consumer_key=consumer_key, consumer_secret=consumer_secret
            )
            self._auth = OAuth2Auth(
                token={"access_token": resp["access_token"], "token_type": "Bearer"}
            )
        # use user auth
        elif all([consumer_key, consumer_secret, access_token, access_secret]):
            self._auth = OAuth1Auth(
                client_id=consumer_key,
                client_secret=consumer_secret,
                token=access_token,
                token_secret=access_secret,
            )
            self.rate_limit = RateLimit("user")
            self.auth_user_id = self.get_uid_from_access_token_key(
                access_token=access_token
            )
        # use oauth flow by hand
        elif consumer_key and consumer_secret and oauth_flow:
            pass
        elif client_id and oauth_flow:
            pass
        else:
            raise PyTwitterError("Need oauth")

    @staticmethod
    def get_uid_from_access_token_key(access_token: str) -> str:
        """
        Get uid from access token, ex: 1323843269210460160-xxx
        :param access_token: Access token
        :return: uid
        """
        uid, _ = access_token.split("-")
        return uid

    def _request(
        self, url, verb="GET", params=None, data=None, json=None, enforce_auth=True
    ) -> Response:
        """
        Request for twitter api url
        :param url: The api location for twitter
        :param verb: HTTP Method, like GET,POST,PUT.
        :param params: The url params to send in the body of the request.
        :param data: The form data to send in the body of the request.
        :param json: The json data to send in the body of the request.
        :param enforce_auth: Whether need auth
        :return: A json object
        """
        auth = None
        if enforce_auth:
            if not self._auth:
                raise PyTwitterError("The twitter.Api instance must be authenticated.")

            auth = self._auth

            if url and self.sleep_on_rate_limit:
                limit = self.rate_limit.get_limit(url=url, method=verb)
                if limit.remaining == 0:
                    s_time = max((limit.reset - time.time()), 0) + 10.0
                    logger.debug(
                        f"Rate limited requesting [{url}], sleeping for [{s_time}]"
                    )
                    time.sleep(s_time)

        resp = self.session.request(
            url=url,
            method=verb,
            params=params,
            data=data,
            auth=auth,
            json=json,
            timeout=self.timeout,
            proxies=self.proxies,
        )

        if url and self.rate_limit:
            self.rate_limit.set_limit(url=url, headers=resp.headers, method=verb)

        return resp

    def _get_oauth1_session(
        self,
        callback_uri: Optional[str] = None,
        **kwargs,
    ) -> OAuth1Session:
        """
        :param callback_uri: The URL that twitter redirect back to after the user logged in.
        :param kwargs: Additional parameters for oauth1.
        :return: OAuth Session
        """
        # check app credentials
        if not (self.consumer_key and self.consumer_secret):
            raise PyTwitterError({"message": "OAuth need your app credentials"})

        if callback_uri is None:
            callback_uri = self.callback_uri

        session = OAuth1Session(
            client_id=self.consumer_key,
            client_secret=self.consumer_secret,
            callback_uri=callback_uri,
            **kwargs,
        )
        return session

    def get_authorize_url(self, callback_uri=None, **kwargs) -> str:
        """
        Get url which to do authorize.
        :param callback_uri: The URL you wish your user to be redirected to.
        :param kwargs: Additional parameters for oauth1, like force_login,screen_name and so on.
        :return: link to authorize
        """
        session = self._get_oauth1_session(callback_uri=callback_uri, **kwargs)
        session.fetch_request_token(self.BASE_REQUEST_TOKEN_URL, proxies=self.proxies)
        return session.create_authorization_url(self.BASE_AUTHORIZE_URL, **kwargs)

    def generate_access_token(
        self, response: str, callback_uri: str = None, **kwargs
    ) -> dict:
        """
        :param response: Response url after user logged in.
        :param callback_uri: The URL you wish your user to be redirected to.
        :param kwargs: Additional parameters for oauth1.
        :return:
        """
        session = self._get_oauth1_session(callback_uri=callback_uri, **kwargs)
        session.parse_authorization_response(response)

        data = session.fetch_access_token(
            self.BASE_ACCESS_TOKEN_URL, proxies=self.proxies
        )
        self._auth = OAuth1Auth(
            client_id=self.consumer_key,
            client_secret=self.consumer_secret,
            token=data["oauth_token"],
            token_secret=data["oauth_token_secret"],
        )
        if "user_id" in data:
            self.auth_user_id = data["user_id"]
        else:
            self.auth_user_id = self.get_uid_from_access_token_key(
                access_token=data["oauth_token"]
            )
        return data

    def invalidate_access_token(self) -> dict:
        """
        Revoke an issued OAuth access_token by presenting its client credentials

        :return:
        """
        if not self._auth:
            raise PyTwitterError("Must have authorized credentials")

        if not isinstance(self._auth, OAuth1Auth):
            raise PyTwitterError("Can only revoke oauth1 token")

        resp = requests.post(
            url="https://api.twitter.com/1.1/oauth/invalidate_token",
        )
        data = self._parse_response(resp=resp)
        return data

    def generate_bearer_token(self, consumer_key: str, consumer_secret: str) -> dict:
        """
        :param consumer_key: Your app consumer key
        :param consumer_secret: Your app consumer secret
        :return: token data
        """
        bearer_token = base64.b64encode(f"{consumer_key}:{consumer_secret}".encode())
        headers = {
            "Authorization": f"Basic {bearer_token.decode()}",
            "Content-Type": "application/x-www-form-urlencoded;charset=UTF-8",
        }
        resp = requests.post(
            url="https://api.twitter.com/oauth2/token",
            data={"grant_type": "client_credentials"},
            headers=headers,
        )
        data = self._parse_response(resp=resp)
        return data

    def invalidate_bearer_token(
        self, consumer_key: str, consumer_secret: str, access_token: str
    ) -> dict:
        """
        Invalidating a Bearer Token

        :param consumer_key: Your app consumer key
        :param consumer_secret: Your app consumer secret
        :param access_token: Token to be invalidated
        :return: token data
        """
        bearer_token = base64.b64encode(f"{consumer_key}:{consumer_secret}".encode())
        headers = {
            "Authorization": f"Basic {bearer_token.decode()}",
            "Content-Type": "application/x-www-form-urlencoded",
        }
        resp = requests.post(
            url="https://api.twitter.com/oauth2/invalidate_token",
            data={"access_token": access_token},
            headers=headers,
        )
        data = self._parse_response(resp=resp)
        return data

    def _get_oauth2_session(
        self,
        redirect_uri: Optional[str] = None,
        scope: Optional[List[str]] = None,
        **kwargs,
    ) -> OAuth2Session:
        """
        :param redirect_uri: The URL that twitter redirect back to after the user logged in.
        :param scope: A list of permission string to request from the user to using your app.
        :param kwargs: Additional parameters for oauth.
        :return: OAuth Session
        """
        # check app credentials
        if not self.client_id:
            raise PyTwitterError({"message": "OAuth need your app credentials"})

        if redirect_uri is None:
            redirect_uri = self.callback_uri
        if scope is None:
            scope = self.scopes

        session = OAuth2Session(
            client_id=self.client_id,
            scope=scope,
            redirect_uri=redirect_uri,
            code_challenge_method="S256",
            **kwargs,
        )
        return session

    def get_oauth2_authorize_url(
        self, redirect_uri: str = None, scope: Optional[List[str]] = None, **kwargs
    ) -> Tuple[str, str, str]:
        """
        :param redirect_uri: The URL that twitter redirect back to after the user logged in.
        :param scope: A list of permission string to request from the user to using your app.
        :param kwargs: Additional parameters for oauth.
        :return: Authorization url, code_verifier, state
        """
        session = self._get_oauth2_session(
            redirect_uri=redirect_uri,
            scope=scope,
            **kwargs,
        )
        code_verifier = base64.urlsafe_b64encode(os.urandom(40)).decode("utf-8")
        code_verifier = re.sub("[^a-zA-Z0-9]+", "", code_verifier)

        authorization_url, state = session.create_authorization_url(
            url=self.BASE_OAUTH2_AUTHORIZE_URL, code_verifier=code_verifier
        )
        return authorization_url, code_verifier, state

    def generate_oauth2_access_token(
        self, response: str, code_verifier: str, redirect_uri: str = None
    ) -> dict:
        """
        :param response: Response url after user logged in.
        :param code_verifier: Code verifier when your
        :param redirect_uri:
        :return:
        """
        session = self._get_oauth2_session(redirect_uri=redirect_uri)

        token = session.fetch_token(
            url=self.BASE_OAUTH2_ACCESS_TOKEN_URL,
            authorization_response=response,
            code_verifier=code_verifier,
        )
        self._auth = OAuth2Auth(token=token["access_token"])
        return token

    @staticmethod
    def _parse_response(resp: Response) -> dict:
        """
        :param resp: Response
        :return: json data
        """
        try:
            data = resp.json()
        except ValueError:
            raise PyTwitterError(f"Unknown error: {resp.content}")

        if not resp.ok:
            raise PyTwitterError(data)

        # note:
        # If only errors will raise
        if "errors" in data and len(data.keys()) == 1:
            raise PyTwitterError(data["errors"])

        # v1 token not
        if "reason" in data:
            raise PyTwitterError(data)

        return data

    def _get(
        self,
        url: str,
        params: Optional[dict],
        cls,
        multi: bool = False,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        :param url: Url for twitter api
        :param params: Parameters for api
        :param cls: Class for the entity
        :param multi: Whether multiple result
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :returns:
            - data: data for the entity like user,tweet...
            - includes: If have expansions, will return
        """
        resp = self._request(url=url, params=params)
        resp_json = self._parse_response(resp)

        if return_json:
            return resp_json
        else:
            data, includes, meta, errors = (
                resp_json.get("data", []),
                resp_json.get("includes"),
                resp_json.get("meta"),
                resp_json.get("errors"),
            )
            if multi:
                data = [cls.new_from_json_dict(item) for item in data]
            else:
                data = cls.new_from_json_dict(data)

            res = md.Response(
                data=data,
                includes=md.Includes.new_from_json_dict(includes),
                meta=md.Meta.new_from_json_dict(meta),
                errors=[md.Error.new_from_json_dict(err) for err in errors]
                if errors is not None
                else None,
            )
            return res

    def get_tweets(
        self,
        tweet_ids: Optional[Union[str, List, Tuple]],
        *,
        expansions: Optional[Union[str, List, Tuple]] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        media_fields: Optional[Union[str, List, Tuple]] = None,
        place_fields: Optional[Union[str, List, Tuple]] = None,
        poll_fields: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns a variety of information about the Tweet specified by the requested ID or list of IDs.

        :param tweet_ids: The IDs for target users, Up to 100 are allowed in a single request.
        :param expansions: Fields for the expansions.
        :param tweet_fields: Fields for the tweet object.
        :param media_fields: Fields for the media object.
        :param place_fields: Fields for the place object.
        :param poll_fields: Fields for the poll object.
        :param user_fields: Fields for the user object.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :returns:
            - data: data for the tweets
            - includes: expansions data.
        """

        args = {
            "ids": enf_comma_separated(name="tweet_ids", value=tweet_ids),
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "media.fields": enf_comma_separated(
                name="media_fields", value=media_fields
            ),
            "place.fields": enf_comma_separated(
                name="place_fields", value=place_fields
            ),
            "poll.fields": enf_comma_separated(name="poll_fields", value=poll_fields),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
        }

        return self._get(
            url=f"{self.BASE_URL_V2}/tweets",
            params=args,
            cls=md.Tweet,
            multi=True,
            return_json=return_json,
        )

    def get_tweet(
        self,
        tweet_id: str,
        *,
        expansions: Optional[Union[str, List, Tuple]] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        media_fields: Optional[Union[str, List, Tuple]] = None,
        place_fields: Optional[Union[str, List, Tuple]] = None,
        poll_fields: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns a variety of information about a single Tweet specified by the requested ID.

        :param tweet_id: The ID of target tweet.
        :param expansions: Fields for the expansions.
        :param tweet_fields: Fields for the tweet object.
        :param media_fields: Fields for the media object.
        :param place_fields: Fields for the place object.
        :param poll_fields: Fields for the poll object.
        :param user_fields: Fields for the user object.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :returns:
            - data: data for the tweet self.
            - includes: expansions data.
        """

        args = {
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "media.fields": enf_comma_separated(
                name="media_fields", value=media_fields
            ),
            "place.fields": enf_comma_separated(
                name="place_fields", value=place_fields
            ),
            "poll.fields": enf_comma_separated(name="poll_fields", value=poll_fields),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
        }
        return self._get(
            url=f"{self.BASE_URL_V2}/tweets/{tweet_id}",
            params=args,
            cls=md.Tweet,
            return_json=return_json,
        )

    def create_tweet(
        self,
        *,
        text: Optional[str] = None,
        direct_message_deep_link: Optional[str] = None,
        for_super_followers_only: Optional[bool] = None,
        geo_place_id: Optional[str] = None,
        media_media_ids: Optional[List[str]] = None,
        media_tagged_user_ids: Optional[List[str]] = None,
        poll_duration_minutes: Optional[int] = None,
        poll_options: Optional[str] = None,
        quote_tweet_id: Optional[str] = None,
        reply_exclude_reply_user_ids: Optional[List[str]] = None,
        reply_in_reply_to_tweet_id: Optional[str] = None,
        reply_settings: Optional[str] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Tweet]:
        """
        Creates a Tweet on behalf of an authenticated user.

        :param text: Text of the Tweet being created. This field is required if media.media_ids is not present.
        :param direct_message_deep_link: Tweets a link directly to a Direct Message conversation with an account.
        :param for_super_followers_only: Allows you to Tweet exclusively for Super Followers.
        :param geo_place_id: Place ID being attached to the Tweet for geo location.
        :param media_media_ids: A list of Media IDs being attached to the Tweet.
            This is only required if the request includes the tagged_user_ids.
        :param media_tagged_user_ids: A list of User IDs being tagged in the Tweet with Media.
            If the user you're tagging doesn't have photo-tagging enabled, their names won't show up
            in the list of tagged users even though the Tweet is successfully created.
        :param poll_duration_minutes: Duration of the poll in minutes for a Tweet with a poll.
            This is only required if the request includes poll.options
        :param poll_options: A list of poll options for a Tweet with a poll.
            For the request to be successful it must also include duration_minutes too.
        :param quote_tweet_id: Link to the Tweet being quoted.
        :param reply_exclude_reply_user_ids: A list of User IDs to be excluded from the reply Tweet
            thus removing a user from a thread.
        :param reply_in_reply_to_tweet_id: Tweet ID of the Tweet being replied to. Please note that in_reply_to_tweet_id
            needs to be in the request if exclude_reply_user_ids is present.
        :param reply_settings: Settings to indicate who can reply to the Tweet.
            Options include "mentionedUsers" and "following". If the field isn’t specified, it will default to everyone.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :returns: Data for tweet created.
        """
        args = {}
        if geo_place_id is not None:
            args["geo"] = {"place_id": geo_place_id}
        if media_media_ids is not None or media_tagged_user_ids is not None:
            args["media"] = {
                "media_ids": media_media_ids,
                "tagged_user_ids": media_tagged_user_ids,
            }
        if poll_duration_minutes is not None or poll_options is not None:
            args["poll"] = {
                "duration_minutes": poll_duration_minutes,
                "options": poll_options,
            }
        if (
            reply_in_reply_to_tweet_id is not None
            or reply_exclude_reply_user_ids is not None
        ):
            args["reply"] = {
                "exclude_reply_user_ids": reply_exclude_reply_user_ids,
                "in_reply_to_tweet_id": reply_in_reply_to_tweet_id,
            }
        if direct_message_deep_link is not None:
            args["direct_message_deep_link"] = direct_message_deep_link
        if for_super_followers_only is not None:
            args["for_super_followers_only"] = for_super_followers_only
        if quote_tweet_id is not None:
            args["quote_tweet_id"] = quote_tweet_id
        if reply_settings is not None:
            args["reply_settings"] = reply_settings
        if text is not None:
            args["text"] = text

        resp = self._request(
            url=f"{self.BASE_URL_V2}/tweets",
            verb="POST",
            json=args,
        )
        data = self._parse_response(resp=resp)
        if return_json:
            return data
        else:
            return md.Tweet.new_from_json_dict(data=data["data"])

    def delete_tweet(
        self,
        tweet_id: str,
    ) -> dict:
        """
        Allows a user or authenticated user ID to delete a Tweet.

        :param tweet_id: ID for the tweet will be deleted.
        :return: Tweet delete status.
        """
        resp = self._request(
            url=f"{self.BASE_URL_V2}/tweets/{tweet_id}",
            verb="DELETE",
        )
        data = self._parse_response(resp=resp)
        return data

    def get_timelines(
        self,
        user_id: str,
        *,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        since_id: Optional[str] = None,
        until_id: Optional[str] = None,
        max_results: Optional[int] = None,
        pagination_token: Optional[str] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        exclude: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        media_fields: Optional[Union[str, List, Tuple]] = None,
        place_fields: Optional[Union[str, List, Tuple]] = None,
        poll_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns Tweets composed by a single user

        :param user_id: The id for target user.
        :param start_time: Oldest or earliest UTC timestamp for tweets, format YYYY-MM-DDTHH:mm:ssZ.
        :param end_time: Newest or most recent UTC timestamp for tweets, format YYYY-MM-DDTHH:mm:ssZ.
        :param since_id: Greater than (that is, more recent than) tweet id for response. Exclude this since_id.
        :param until_id: Less than (that is, older than) tweet id for response. Exclude this until_id.
        :param max_results: The maximum number of results to be returned per page. Number between 5 and the 100.
        By default, each page will return 10 results.
        :param pagination_token: Token for the pagination.
        :param tweet_fields: Fields for the tweet object.
        :param exclude: Fields for types of Tweets to exclude from the response.
        :param expansions: Fields for the expansions.
        :param user_fields: Fields for the user object, Expansion required.
        :param media_fields: Fields for the media object, Expansion required.
        :param place_fields: Fields for the place object, Expansion required.
        :param poll_fields: Fields for the poll object, Expansion required.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return: Response instance or json.
        """

        args = {
            "start_time": start_time,
            "end_time": end_time,
            "since_id": since_id,
            "until_id": until_id,
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "exclude": enf_comma_separated(name="exclude", value=exclude),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "media.fields": enf_comma_separated(
                name="media_fields", value=media_fields
            ),
            "place.fields": enf_comma_separated(
                name="place_fields", value=place_fields
            ),
            "poll.fields": enf_comma_separated(name="poll_fields", value=poll_fields),
            "max_results": max_results,
            "pagination_token": pagination_token,
        }
        return self._get(
            url=f"{self.BASE_URL_V2}/users/{user_id}/tweets",
            params=args,
            cls=md.Tweet,
            multi=True,
            return_json=return_json,
        )

    def get_mentions(
        self,
        user_id: str,
        *,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        since_id: Optional[str] = None,
        until_id: Optional[str] = None,
        max_results: Optional[int] = None,
        pagination_token: Optional[str] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        media_fields: Optional[Union[str, List, Tuple]] = None,
        place_fields: Optional[Union[str, List, Tuple]] = None,
        poll_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns Tweets mentioning user specified by ID.

        :param user_id: The id for target user.
        :param start_time: Oldest or earliest UTC timestamp for tweets, format YYYY-MM-DDTHH:mm:ssZ.
        :param end_time: Newest or most recent UTC timestamp for tweets, format YYYY-MM-DDTHH:mm:ssZ.
        :param since_id: Greater than (that is, more recent than) tweet id for response. Exclude this since_id.
        :param until_id: Less than (that is, older than) tweet id for response. Exclude this until_id.
        :param max_results: The maximum number of results to be returned per page. Number between 5 and the 100.
        By default, each page will return 10 results.
        :param pagination_token: Token for the pagination.
        :param tweet_fields: Fields for the tweet object.
        :param expansions: Fields for the expansions.
        :param user_fields: Fields for the user object, Expansion required.
        :param media_fields: Fields for the media object, Expansion required.
        :param place_fields: Fields for the place object, Expansion required.
        :param poll_fields: Fields for the poll object, Expansion required.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return: Response instance or json.
        """
        args = {
            "start_time": start_time,
            "end_time": end_time,
            "since_id": since_id,
            "until_id": until_id,
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "media.fields": enf_comma_separated(
                name="media_fields", value=media_fields
            ),
            "place.fields": enf_comma_separated(
                name="place_fields", value=place_fields
            ),
            "poll.fields": enf_comma_separated(name="poll_fields", value=poll_fields),
            "max_results": max_results,
            "pagination_token": pagination_token,
        }
        return self._get(
            url=f"{self.BASE_URL_V2}/users/{user_id}/mentions",
            params=args,
            cls=md.Tweet,
            multi=True,
            return_json=return_json,
        )

    def search_tweets(
        self,
        query: str,
        query_type: str = "recent",
        *,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        since_id: Optional[str] = None,
        until_id: Optional[str] = None,
        max_results: Optional[int] = None,
        next_token: Optional[str] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        media_fields: Optional[Union[str, List, Tuple]] = None,
        place_fields: Optional[Union[str, List, Tuple]] = None,
        poll_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Search tweets endpoint has two type:
            - recent (default): Returns Tweets from the last seven days that match a search query.
            - all: Returns the complete history of public Tweets matching a search query;
            since the first Tweet was created March 26, 2006.
            But this type only for who have been approved for the `Academic Research product track`.

        :param query: One rule for matching Tweets.
        :param query_type: Accepted values: recent or all
        :param start_time: Oldest or earliest UTC timestamp for tweets, format YYYY-MM-DDTHH:mm:ssZ.
        :param end_time: Newest or most recent UTC timestamp for tweets, format YYYY-MM-DDTHH:mm:ssZ.
        :param since_id: Greater than (that is, more recent than) tweet id for response. Exclude this since_id.
        :param until_id: Less than (that is, older than) tweet id for response. Exclude this until_id.
        :param max_results: The maximum number of results to be returned per page. Number between 10 and up to 500.
        By default, each page will return 10 results.
        :param next_token: Token for the pagination.
        :param tweet_fields: Fields for the tweet object.
        :param expansions: Fields for the expansions.
        :param user_fields: Fields for the user object, Expansion required.
        :param media_fields: Fields for the media object, Expansion required.
        :param place_fields: Fields for the place object, Expansion required.
        :param poll_fields: Fields for the poll object, Expansion required.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return: Response instance or json.
        """

        args = {
            "query": query,
            "start_time": start_time,
            "end_time": end_time,
            "since_id": since_id,
            "until_id": until_id,
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "media.fields": enf_comma_separated(
                name="media_fields", value=media_fields
            ),
            "place.fields": enf_comma_separated(
                name="place_fields", value=place_fields
            ),
            "poll.fields": enf_comma_separated(name="poll_fields", value=poll_fields),
            "max_results": max_results,
            "next_token": next_token,
        }

        if query_type == "recent":
            url = f"{self.BASE_URL_V2}/tweets/search/recent"
        elif query_type == "all":
            url = f"{self.BASE_URL_V2}/tweets/search/all"
        else:
            raise PyTwitterError(f"Not support for query type: {query_type}")

        return self._get(
            url=url,
            params=args,
            cls=md.Tweet,
            multi=True,
            return_json=return_json,
        )

    def get_tweets_counts(
        self,
        query: str,
        search_type: str = "recent",
        *,
        granularity: Optional[str] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        since_id: Optional[str] = None,
        until_id: Optional[str] = None,
        next_token: Optional[str] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Get count of Tweets that match a search query.

        :param query: One rule for matching Tweets.
            - Standard Project can use the basic set of operators and can make queries up to 512 characters long
            - Academic Research Project can use all available operators and can make queries up to 1,024 characters long
        :param search_type: Accepted values:
            - recent: For all apps, Only can get tweets from the last seven days.
            - all: only for Academic Research product track. Since the first Tweet was created March 26, 2006.
        :param granularity: The timeseries count data to be grouped by. Can be day, hour or minute
        :param start_time: The oldest UTC timestamp from which the Tweets will be provided.
        :param end_time: The newest, most recent UTC timestamp to which the Tweets will be provided.
        :param since_id: Returns results with a Tweet ID greater than (for example, more recent than) the specified ID.
        :param until_id: Returns results with a Tweet ID less than (that is, older than) the specified ID.
        :param next_token: This parameter is used to get the next 'page' of results.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: Data for the counts.
            - meta: Meta data for request.
        """

        args = {
            "query": query,
            "granularity": granularity,
            "start_time": start_time,
            "end_time": end_time,
            "since_id": since_id,
            "until_id": until_id,
            "next_token": next_token,
        }

        if search_type == "recent":
            url = f"{self.BASE_URL_V2}/tweets/counts/recent"
        elif search_type == "all":
            url = f"{self.BASE_URL_V2}/tweets/counts/all"
        else:
            raise PyTwitterError(f"Not support type for {search_type}")

        return self._get(
            url=url,
            params=args,
            cls=md.TweetCount,
            multi=True,
            return_json=return_json,
        )

    def get_tweet_retweeted_users(
        self,
        tweet_id: str,
        *,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Get information about who has Retweeted a Tweet.

        :param tweet_id: The tweet ID whose retweeted users you would like to retrieve.
        :param user_fields: Fields for the user object.
        :param expansions: Fields for the expansions now only `pinned_tweet_id`.
        :param tweet_fields: Fields for the tweet object, Expansions required.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the users.
            - includes: expansions data.
            - meta: pagination details
        """
        args = {
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
        }
        return self._get(
            url=f"{self.BASE_URL_V2}/tweets/{tweet_id}/retweeted_by",
            params=args,
            cls=md.User,
            multi=True,
            return_json=return_json,
        )

    def retweet_tweet(self, user_id: str, tweet_id: str) -> dict:
        """
        Allows user to retweet a tweet.
        :param user_id: The user ID who you are retweeting a Tweet on behalf of.
                It must match your user ID which authorize with the access token.
        :param tweet_id: The ID of the Tweet that you would retweet.
        :return: retweet status data
        """

        resp = self._request(
            url=f"{self.BASE_URL_V2}/users/{user_id}/retweets",
            verb="POST",
            json={"tweet_id": tweet_id},
        )
        data = self._parse_response(resp=resp)
        return data

    def remove_retweet_tweet(self, user_id: str, tweet_id: str) -> dict:
        """
        Allows a user to remove the Retweet of a Tweet.

        :param user_id: The user ID who you are removing a Retweet of a Tweet on behalf of.
                It must match your user ID which authorize with the access token.
        :param tweet_id: The ID of the Tweet that you would remove retweet status.
        :return: retweet status data
        """

        resp = self._request(
            url=f"{self.BASE_URL_V2}/users/{user_id}/retweets/{tweet_id}", verb="DELETE"
        )
        data = self._parse_response(resp=resp)
        return data

    def get_tweet_liking_users(
        self,
        tweet_id: str,
        *,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Get information about a Tweet’s liking users.

        :param tweet_id: The tweet ID whose liking users you would like to retrieve.
        :param expansions: Fields for the expansions.
        By default, each page will return 100 results.
        :param tweet_fields: Fields for the tweet object.
        :param user_fields: Fields for the user object, Expansion required.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the users.
            - includes: expansions data.
            - meta: pagination details
        """
        args = {
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
        }
        return self._get(
            url=f"{self.BASE_URL_V2}/tweets/{tweet_id}/liking_users",
            params=args,
            cls=md.User,
            multi=True,
            return_json=return_json,
        )

    def get_user_liked_tweets(
        self,
        user_id: str,
        *,
        pagination_token: Optional[str] = None,
        max_results: Optional[int] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        media_fields: Optional[Union[str, List, Tuple]] = None,
        place_fields: Optional[Union[str, List, Tuple]] = None,
        poll_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Get information about a user’s liked Tweets.

        :param user_id: The user ID whose liked tweets you would like to retrieve.
        :param expansions: Fields for the expansions.
        :param pagination_token: Token for the pagination.
        :param max_results: The maximum number of results to be returned per page. Number between 1 and the 1000.
        By default, each page will return 100 results.
        :param tweet_fields: Fields for the tweet object.
        :param user_fields: Fields for the user object, Expansion required.
        :param media_fields: Fields for the media object, Expansion required.
        :param place_fields: Fields for the place object, Expansion required.
        :param poll_fields: Fields for the poll object, Expansion required.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the tweets.
            - includes: expansions data.
            - meta: pagination details
        """
        args = {
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "media.fields": enf_comma_separated(
                name="media_fields", value=media_fields
            ),
            "place.fields": enf_comma_separated(
                name="place_fields", value=place_fields
            ),
            "poll.fields": enf_comma_separated(name="poll_fields", value=poll_fields),
            "max_results": max_results,
            "pagination_token": pagination_token,
        }
        return self._get(
            url=f"{self.BASE_URL_V2}/users/{user_id}/liked_tweets",
            params=args,
            cls=md.Tweet,
            multi=True,
            return_json=return_json,
        )

    def like_tweet(self, user_id: str, tweet_id: str) -> dict:
        """
        Allows user to like tweet.

        :param user_id: The user ID who you are liking a Tweet on behalf of.
                It must match your user ID which authorize with the access token.
        :param tweet_id: The ID of the Tweet that you would like.
        :return: like status data
        """

        resp = self._request(
            url=f"{self.BASE_URL_V2}/users/{user_id}/likes",
            verb="POST",
            json={"tweet_id": tweet_id},
        )
        data = self._parse_response(resp=resp)
        return data

    def unlike_tweet(self, user_id: str, tweet_id: str) -> dict:
        """
        Allows user to remove like status from a tweet.

        :param user_id: The user ID who you are removing a Like of a Tweet on behalf of.
                It must match your user ID which authorize with the access token.
        :param tweet_id: The ID of the Tweet that you would remove like status.
        :return: like status data
        """

        resp = self._request(
            url=f"{self.BASE_URL_V2}/users/{user_id}/likes/{tweet_id}", verb="DELETE"
        )
        data = self._parse_response(resp=resp)
        return data

    def hidden_reply(self, tweet_id: str, hidden: Optional[bool] = True) -> dict:
        """
        Hide or un-hide a reply to a Tweet.

        Note: This api must with OAuth 1.0a User context.

        :param tweet_id: ID of the tweet to hide or un-hide,
        :param hidden: If set True, will hide reply, If set False, will un-hide reply. Default is True.
        :return: status for hide or un-hide.
        """

        resp = self._request(
            url=f"{self.BASE_URL_V2}/tweets/{tweet_id}/hidden",
            verb="PUT",
            json={"hidden": hidden},
        )
        data = self._parse_response(resp=resp)
        return data

    def get_users(
        self,
        *,
        ids: Optional[Union[str, List, Tuple]] = None,
        usernames: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns a variety of information about one or more users specified by the requested IDs or usernames.

        :param ids: The IDs for target users, Up to 100 are allowed in a single request.
        :param usernames: The username for target users, Up to 100 are allowed in a single request.
            Either ids or username is required for this method.
        :param user_fields: Fields for the user object.
        :param expansions: Fields for expansions.
        :param tweet_fields: Fields for the tweet object.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :returns:
            - data: data for the users
            - includes: expansions data.
        """
        args = {
            "ids": enf_comma_separated(name="ids", value=ids),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
        }

        if ids:
            args["ids"] = enf_comma_separated(name="ids", value=ids)
            path = "users"
        elif usernames:
            args["usernames"] = enf_comma_separated(name="usernames", value=usernames)
            path = "users/by"
        else:
            raise PyTwitterError("Specify at least one of ids or usernames")

        return self._get(
            url=f"{self.BASE_URL_V2}/{path}",
            params=args,
            cls=md.User,
            multi=True,
            return_json=return_json,
        )

    def get_user(
        self,
        *,
        user_id: Optional[str] = None,
        username: Optional[str] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns a variety of information about a single user specified by the requested ID or username.

        :param user_id: The ID of target user.
        :param username: The username of target user.
        :param user_fields: Fields for the user object.
        :param expansions: Fields for expansions.
        :param tweet_fields: Fields for the tweet object.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :returns:
            - data: data for the user
            - includes: expansions data.
        """

        args = {
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
        }

        if user_id:
            path = f"users/{user_id}"
        elif username:
            path = f"users/by/username/{username}"
        else:
            raise PyTwitterError("Specify at least one of user_id or username")

        return self._get(
            url=f"{self.BASE_URL_V2}/{path}",
            params=args,
            cls=md.User,
            return_json=return_json,
        )

    def get_me(
        self,
        *,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ):
        """
        Returns information about an authorized user.

        :param user_fields: Fields for the user object.
        :param expansions: Fields for expansions.
        :param tweet_fields: Fields for the tweet object.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :returns:
            - data: data for the user
            - includes: expansions data.
        """
        args = {
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
        }

        return self._get(
            url=f"{self.BASE_URL_V2}/users/me",
            params=args,
            cls=md.User,
            return_json=return_json,
        )

    def get_following(
        self,
        user_id: str,
        *,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        max_results: Optional[int] = None,
        pagination_token: Optional[str] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns a list of users the specified user ID is following.

        :param user_id: The user ID whose following you would like to retrieve.
        :param expansions: Fields for the expansions.
        :param user_fields: Fields for the user object.
        :param tweet_fields: Fields for the tweet object.
        :param max_results: The maximum number of results to be returned per page. Number between 1 and the 1000.
        By default, each page will return 100 results.
        :param pagination_token: Token for the pagination.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the following.
            - includes: expansions data.
            - meta: pagination details
        """

        args = {
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "max_results": max_results,
            "pagination_token": pagination_token,
        }

        return self._get(
            url=f"{self.BASE_URL_V2}/users/{user_id}/following",
            params=args,
            cls=md.User,
            multi=True,
            return_json=return_json,
        )

    def get_followers(
        self,
        user_id: str,
        *,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        max_results: Optional[int] = None,
        pagination_token: Optional[str] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns a list of users who are followers of the specified user ID.

        :param user_id: The user ID whose following you would like to retrieve.
        :param expansions: Fields for the expansions.
        :param user_fields: Fields for the user object.
        :param tweet_fields: Fields for the tweet object.
        :param max_results: The maximum number of results to be returned per page. Number between 1 and the 1000.
        By default, each page will return 100 results.
        :param pagination_token: Token for the pagination.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the following.
            - includes: expansions data.
            - meta: pagination details
        """
        args = {
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "max_results": max_results,
            "pagination_token": pagination_token,
        }

        return self._get(
            url=f"{self.BASE_URL_V2}/users/{user_id}/followers",
            params=args,
            cls=md.User,
            multi=True,
            return_json=return_json,
        )

    def follow_user(self, user_id: str, target_user_id: str) -> dict:
        """
        Allows a user ID to follow another user.
        If the target user does not have public Tweets, this endpoint will send a follow request.

        :param user_id: The user ID who you would like to initiate the follow on behalf of.
                        It must match the authenticating user.
        :param target_user_id: The target user ID of user to follow
        :return: follow status data
        """

        resp = self._request(
            url=f"{self.BASE_URL_V2}/users/{user_id}/following",
            verb="POST",
            json={"target_user_id": target_user_id},
        )
        data = self._parse_response(resp)
        return data

    def unfollow_user(self, user_id: str, target_user_id: str) -> dict:
        """
        Allows a user ID to unfollow another user.

        :param user_id: The user ID who you would like to initiate the unfollow on behalf of.
                        It must match the username of the authenticating user.
        :param target_user_id: The user ID of user to unfollow.
        :return: follow status data
        """
        resp = self._request(
            url=f"{self.BASE_URL_V2}/users/{user_id}/following/{target_user_id}",
            verb="DELETE",
        )
        data = self._parse_response(resp)
        return data

    def get_blocking_users(
        self,
        user_id: str,
        *,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        max_results: Optional[int] = None,
        pagination_token: Optional[str] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns a list of users who are blocked by the specified user ID.

        :param user_id: The user ID whose blocking you would like to retrieve.
        :param expansions: Fields for the expansions.
        :param user_fields: Fields for the user object.
        :param tweet_fields: Fields for the tweet object.
        :param max_results: The maximum number of results to be returned per page. Number between 1 and the 1000.
        By default, each page will return 100 results.
        :param pagination_token: Token for the pagination.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the blocking.
            - includes: expansions data.
            - meta: pagination details
        """
        args = {
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "max_results": max_results,
            "pagination_token": pagination_token,
        }
        return self._get(
            url=f"{self.BASE_URL_V2}/users/{user_id}/blocking",
            params=args,
            cls=md.User,
            multi=True,
            return_json=return_json,
        )

    def block_user(self, user_id: str, target_user_id: str) -> dict:
        """
        Allows user to block target user.

        :param user_id: The user ID who you would like to initiate the block on behalf of.
                It must match your user ID which authorize with the access token.
        :param target_user_id: The target user ID of user to block
        :return: block status data
        """

        resp = self._request(
            url=f"{self.BASE_URL_V2}/users/{user_id}/blocking",
            verb="POST",
            json={"target_user_id": target_user_id},
        )
        data = self._parse_response(resp)
        return data

    def unblock_user(self, user_id: str, target_user_id: str) -> dict:
        """
        Allows user to unblock another user.

        :param user_id: The user ID who you would like to initiate an unblock on behalf of.
                It must match your user ID which authorize with the access token.
        :param target_user_id: The target user ID of user to block
        :return: delete block status data
        """

        resp = self._request(
            url=f"{self.BASE_URL_V2}/users/{user_id}/blocking/{target_user_id}",
            verb="DELETE",
        )
        data = self._parse_response(resp)
        return data

    def get_user_muting(
        self,
        user_id: str,
        *,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        max_results: Optional[int] = None,
        pagination_token: Optional[str] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns a list of users who are muted by the specified user ID.

        :param user_id: ID for user which you want to get muting.
        :param user_fields: Fields for the user object.
        :param expansions: Fields for the expansions now only `pinned_tweet_id`.
        :param tweet_fields: Fields for the tweet object, Expansions required.
        :param max_results: The maximum number of results to be returned per page. Number between 1 and the 1000.
        :param pagination_token: Token for the pagination.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the users.
            - includes: expansions data.
            - meta: pagination details
        """

        args = {
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "max_results": max_results,
            "pagination_token": pagination_token,
        }
        return self._get(
            url=f"{self.BASE_URL_V2}/users/{user_id}/muting",
            params=args,
            cls=md.User,
            multi=True,
            return_json=return_json,
        )

    def mute_user(self, user_id: str, target_user_id: str) -> dict:
        """
        Allows user to mute the target user.

        :param user_id: The user ID who you would like to initiate the mute on behalf of.
            It must match your own user ID or that of an authenticating user
        :param target_user_id: The user ID of the user that you would like the id to mute.
        :return: Mute status data
        """
        resp = self._request(
            url=f"{self.BASE_URL_V2}/users/{user_id}/muting",
            verb="POST",
            json={"target_user_id": target_user_id},
        )
        data = self._parse_response(resp)
        return data

    def unmute_user(self, user_id: str, target_user_id: str) -> dict:
        """
        Allows user to unmute the target user.

        :param user_id: The user ID who you would like to initiate an unmute on behalf of.
            It must match your own user ID or that of an authenticating user
        :param target_user_id: The user ID of the user that you would like to unmute.
        :return: Unmute status data
        """
        resp = self._request(
            url=f"{self.BASE_URL_V2}/users/{user_id}/muting/{target_user_id}",
            verb="DELETE",
        )
        data = self._parse_response(resp)
        return data

    def get_space(
        self,
        space_id: str,
        *,
        space_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns a variety of information about a single Space specified by the requested ID.

        :param space_id: The ID for the target space.
        :param space_fields: Fields for the space object.
        :param expansions: Fields for expansions.
        :param user_fields: Fields for the user object.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the space
            - includes: expansions data.
        """

        args = {
            "space.fields": enf_comma_separated(
                name="space_fields",
                value=space_fields,
            ),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
        }

        return self._get(
            url=f"{self.BASE_URL_V2}/spaces/{space_id}",
            params=args,
            cls=md.Space,
            return_json=return_json,
        )

    def get_spaces(
        self,
        space_ids: Union[str, List, Tuple],
        *,
        space_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns details for multiple Spaces. Up to 100 comma-separated Spaces IDs can be looked up using this endpoint.

        :param space_ids: The IDs for target spaces, Up to 100 are allowed in a single request.
        :param space_fields: Fields for the space object.
        :param expansions: Fields for expansions.
        :param user_fields: Fields for the user object.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the spaces
            - includes: expansions data.
        """
        args = {
            "ids": enf_comma_separated(name="space_ids", value=space_ids),
            "space.fields": enf_comma_separated(
                name="space_fields",
                value=space_fields,
            ),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
        }

        return self._get(
            url=f"{self.BASE_URL_V2}/spaces",
            params=args,
            cls=md.Space,
            multi=True,
            return_json=return_json,
        )

    def get_spaces_by_creator(
        self,
        creator_ids: Union[str, List, Tuple],
        *,
        space_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        max_results: Optional[int] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns live or scheduled Spaces created by the specified user IDs.
        Up to 100 comma-separated IDs can be looked up using this endpoint.

        :param creator_ids: IDs for the creators, Up to 100 are allowed in a single request.
        :param space_fields: Fields for the space object.
        :param expansions: Fields for expansions.
        :param user_fields: Fields for the user object.
        :param max_results: The maximum number of results to be returned per page. Number between 1 and the 100.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the spaces
            - includes: expansions data.
        """
        args = {
            "user_ids": enf_comma_separated(name="creator_ids", value=creator_ids),
            "space.fields": enf_comma_separated(
                name="space_fields",
                value=space_fields,
            ),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "max_results": max_results,
        }

        return self._get(
            url=f"{self.BASE_URL_V2}/spaces/by/creator_ids",
            params=args,
            cls=md.Space,
            multi=True,
            return_json=return_json,
        )

    def search_spaces(
        self,
        query: str,
        state: str,
        *,
        space_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        max_results: Optional[int] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Return live or scheduled Spaces matching your specified search terms

        :param query: Your search term.
            This can be any text (including mentions and Hashtags) present in the title of the Space.
        :param state: Determines the type of results to return.
            Accepted values: live,scheduled
        :param space_fields: Fields for the space object.
        :param expansions: Fields for expansions.
        :param user_fields: Fields for the user object.
        :param max_results: The maximum number of results to be returned per page. Number between 1 and the 100.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the spaces
            - includes: expansions data.
        """

        args = {
            "query": query,
            "state": state,
            "space.fields": enf_comma_separated(
                name="space_fields",
                value=space_fields,
            ),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "max_results": max_results,
        }

        return self._get(
            url=f"{self.BASE_URL_V2}/spaces/search",
            params=args,
            cls=md.Space,
            multi=True,
            return_json=return_json,
        )

    def get_list(
        self,
        list_id: str,
        *,
        list_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns the details of a specified List.

        :param list_id: ID for the list.
        :param list_fields: Fields for the list object.
        :param expansions: Fields for expansions.
        :param user_fields: Fields for the user object.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the list
            - includes: expansions data.
        """

        args = {
            "list.fields": enf_comma_separated(name="list_fields", value=list_fields),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
        }

        return self._get(
            url=f"{self.BASE_URL_V2}/lists/{list_id}",
            params=args,
            cls=md.TwitterList,
            return_json=return_json,
        )

    def get_user_owned_lists(
        self,
        user_id: str,
        *,
        list_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        max_results: Optional[int] = None,
        pagination_token: Optional[str] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns all Lists owned by the specified user.

        :param user_id: ID for the user.
        :param list_fields: Fields for the list object.
        :param expansions: Fields for expansions.
        :param user_fields: Fields for the user object. Expansion required.
        :param max_results: The maximum number of results to be returned per page. Number between 1 and the 100.
            By default, each page will return 100 results.
        :param pagination_token: Token for the pagination.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data for the lists
            - includes: expansions data.
            - meta: pagination details
        """

        args = {
            "list.fields": enf_comma_separated(name="list_fields", value=list_fields),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "max_results": max_results,
            "pagination_token": pagination_token,
        }
        return self._get(
            url=f"{self.BASE_URL_V2}/users/{user_id}/owned_lists",
            params=args,
            cls=md.TwitterList,
            multi=True,
            return_json=return_json,
        )

    def create_list(
        self,
        name: str,
        *,
        description: Optional[str] = None,
        private: Optional[bool] = None,
        return_json: bool = False,
    ) -> Union[md.TwitterList, dict]:
        """
        Enables the authenticated user to create a List.

        :param name: The name of the List you wish to create.
        :param description: Description of the List.
        :param private: Determine whether the List should be private.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return: Twitter List information.
        """

        args = {"name": name}
        if description is not None:
            args["description"] = description
        if private is not None:
            args["private"] = private

        resp = self._request(
            url=f"{self.BASE_URL_V2}/lists",
            verb="POST",
            json=args,
        )
        data = self._parse_response(resp=resp)
        if return_json:
            return data
        else:
            return md.TwitterList.new_from_json_dict(data["data"])

    def update_list(
        self,
        list_id: str,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
        private: Optional[bool] = None,
    ) -> dict:
        """
        Enables the authenticated user to update the meta data of a specified List that they own.

        :param list_id: The ID of the List to be updated.
        :param name: Updates the name of the List.
        :param description: Updates the description of the List.
        :param private: Determines whether the List should be private.
        :return: Status for update twitter list.
        """

        args = {}
        if name is not None:
            args["name"] = name
        if description is not None:
            args["description"] = description
        if private is not None:
            args["private"] = private

        resp = self._request(
            url=f"{self.BASE_URL_V2}/lists/{list_id}",
            verb="PUT",
            json=args,
        )
        data = self._parse_response(resp=resp)
        return data

    def delete_list(self, list_id: str) -> dict:
        """
        Enables the authenticated user to delete a List that they own.

        :param list_id: The ID of the List to be deleted.
        :return: Status for delete list.
        """
        resp = self._request(
            url=f"{self.BASE_URL_V2}/lists/{list_id}",
            verb="DELETE",
        )
        data = self._parse_response(resp=resp)
        return data

    def get_list_tweets(
        self,
        list_id: str,
        *,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        pagination_token: Optional[str] = None,
        max_results: Optional[int] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns a list of Tweets from the specified List.

        :param list_id: ID for the list
        :param tweet_fields: Fields for the tweet object.
        :param expansions: Fields for the expansions.
        :param user_fields: Fields for the user object, Expansion required.
        :param max_results: The maximum number of results to be returned per page. Number between 1 and the 100.
            By default, each page will return 100 results.
        :param pagination_token: Token for the pagination.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the tweets.
            - includes: expansions data.
            - meta: pagination details
        """

        args = {
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "max_results": max_results,
            "pagination_token": pagination_token,
        }

        return self._get(
            url=f"{self.BASE_URL_V2}/lists/{list_id}/tweets",
            params=args,
            cls=md.Tweet,
            multi=True,
            return_json=return_json,
        )

    def get_list_members(
        self,
        list_id: str,
        *,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        max_results: Optional[int] = None,
        pagination_token: Optional[str] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns a list of users who are members of the specified List.

        :param list_id: ID for the list.
        :param user_fields: Fields for the user object.
        :param expansions: Fields for expansions.
        :param tweet_fields: Fields for the tweet object.
        :param max_results: The maximum number of results to be returned per page. Number between 1 and the 100.
            By default, each page will return 100 results.
        :param pagination_token: Token for the pagination.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the tweets.
            - includes: expansions data.
            - meta: pagination details
        """

        args = {
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "max_results": max_results,
            "pagination_token": pagination_token,
        }

        return self._get(
            url=f"{self.BASE_URL_V2}/lists/{list_id}/members",
            params=args,
            cls=md.User,
            multi=True,
            return_json=return_json,
        )

    def get_user_memberships_lists(
        self,
        user_id: str,
        *,
        list_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        max_results: Optional[int] = None,
        pagination_token: Optional[str] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns all Lists a specified user is a member of.

        :param user_id: ID for the user.
        :param list_fields: Fields for the list object.
        :param expansions: Fields for expansions.
        :param user_fields: Fields for the user object. Expansion required.
        :param max_results: The maximum number of results to be returned per page. Number between 1 and the 100.
            By default, each page will return 100 results.
        :param pagination_token: Token for the pagination.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data for the lists.
            - includes: expansions data.
            - meta: pagination details
        """

        args = {
            "list.fields": enf_comma_separated(name="list_fields", value=list_fields),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "max_results": max_results,
            "pagination_token": pagination_token,
        }

        return self._get(
            url=f"{self.BASE_URL_V2}/users/{user_id}/list_memberships",
            params=args,
            cls=md.TwitterList,
            multi=True,
            return_json=return_json,
        )

    def add_list_member(
        self,
        *,
        list_id: str,
        user_id: str,
    ) -> dict:
        """
        Enables the authenticated user to add a member to a List they own.

        :param list_id: The ID of the List you are adding a member to.
        :param user_id: The ID of the user you wish to add as a member of the List.
        :return: Member added status
        """
        resp = self._request(
            url=f"{self.BASE_URL_V2}/lists/{list_id}/members",
            verb="POST",
            json={"user_id": user_id},
        )
        data = self._parse_response(resp=resp)
        return data

    def remove_list_member(
        self,
        *,
        list_id: str,
        user_id: str,
    ) -> dict:
        """
        Enables the authenticated user to remove a member from a List they own.

        :param list_id: The ID of the List you are removing a member from.
        :param user_id: The ID of the user you wish to remove as a member of the List.
        :return: Member remove status
        """
        resp = self._request(
            url=f"{self.BASE_URL_V2}/lists/{list_id}/members/{user_id}",
            verb="DELETE",
        )
        data = self._parse_response(resp=resp)
        return data

    def follow_list(
        self,
        *,
        user_id: str,
        list_id: str,
    ) -> dict:
        """
        Enables the authenticated user to follow a List.

        :param user_id: The user ID who you are following a List on behalf of, with access token associated.
        :param list_id: The ID of the List that you would like the user id to follow.
        :return: Follow list status
        """
        resp = self._request(
            url=f"{self.BASE_URL_V2}/users/{user_id}/followed_lists",
            verb="POST",
            json={"list_id": list_id},
        )
        data = self._parse_response(resp=resp)
        return data

    def unfollow_list(
        self,
        *,
        user_id: str,
        list_id: str,
    ) -> dict:
        """
        Enables the authenticated user to unfollow a List.

        :param user_id: The user ID who you are unfollowing a List on behalf of, with access token associated.
        :param list_id: The ID of the List that you would like the user id to unfollow.
        :return: Unfollow list status.
        """
        resp = self._request(
            url=f"{self.BASE_URL_V2}/users/{user_id}/followed_lists/{list_id}",
            verb="DELETE",
        )
        data = self._parse_response(resp=resp)
        return data

    def get_list_followers(
        self,
        list_id: str,
        *,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        tweet_fields: Optional[Union[str, List, Tuple]] = None,
        max_results: Optional[int] = None,
        pagination_token: Optional[str] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns a list of users who are followers of the specified List.

        :param list_id: ID for the list.
        :param user_fields: Fields for the user object.
        :param expansions: Fields for expansions.
        :param tweet_fields: Fields for the tweet object.
        :param max_results: The maximum number of results to be returned per page. Number between 1 and the 100.
            By default, each page will return 100 results.
        :param pagination_token: Token for the pagination.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the tweets.
            - includes: expansions data.
            - meta: pagination details
        """

        args = {
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "tweet.fields": enf_comma_separated(
                name="tweet_fields", value=tweet_fields
            ),
            "max_results": max_results,
            "pagination_token": pagination_token,
        }

        return self._get(
            url=f"{self.BASE_URL_V2}/lists/{list_id}/followers",
            params=args,
            cls=md.User,
            multi=True,
            return_json=return_json,
        )

    def get_user_followed_lists(
        self,
        user_id: str,
        *,
        list_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        max_results: Optional[int] = None,
        pagination_token: Optional[str] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns all Lists a specified user follows.

        :param user_id: ID for the user.
        :param list_fields: Fields for the list object.
        :param expansions: Fields for expansions.
        :param user_fields: Fields for the user object. Expansion required.
        :param max_results: The maximum number of results to be returned per page. Number between 1 and the 100.
            By default, each page will return 100 results.
        :param pagination_token: Token for the pagination.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data for the lists.
            - includes: expansions data.
            - meta: pagination details
        """

        args = {
            "list.fields": enf_comma_separated(name="list_fields", value=list_fields),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
            "max_results": max_results,
            "pagination_token": pagination_token,
        }

        return self._get(
            url=f"{self.BASE_URL_V2}/users/{user_id}/followed_lists",
            params=args,
            cls=md.TwitterList,
            multi=True,
            return_json=return_json,
        )

    def get_user_pinned_lists(
        self,
        user_id: str,
        *,
        list_fields: Optional[Union[str, List, Tuple]] = None,
        expansions: Optional[Union[str, List, Tuple]] = None,
        user_fields: Optional[Union[str, List, Tuple]] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns the Lists pinned by a specified user.

        :param user_id: ID for the user.
        :param list_fields: Fields for the list object.
        :param expansions: Fields for expansions.
        :param user_fields: Fields for the user object. Expansion required.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data for the lists
            - includes: expansions data.
            - meta: pagination details
        """
        args = {
            "list.fields": enf_comma_separated(name="list_fields", value=list_fields),
            "expansions": enf_comma_separated(name="expansions", value=expansions),
            "user.fields": enf_comma_separated(name="user_fields", value=user_fields),
        }
        return self._get(
            url=f"{self.BASE_URL_V2}/users/{user_id}/pinned_lists",
            params=args,
            cls=md.TwitterList,
            multi=True,
            return_json=return_json,
        )

    def pin_list(
        self,
        *,
        user_id: str,
        list_id: str,
    ) -> dict:
        """
        Enables the authenticated user to pin a List.

        :param user_id: The user ID who you are pinning a List on behalf of, with access token associated.
        :param list_id: The ID of the List that you would like the user id to pin.
        :return: Pin list status.
        """
        resp = self._request(
            url=f"{self.BASE_URL_V2}/users/{user_id}/pinned_lists",
            verb="POST",
            json={"list_id": list_id},
        )
        data = self._parse_response(resp=resp)
        return data

    def unpin_list(
        self,
        *,
        user_id: str,
        list_id: str,
    ) -> dict:
        """
        Enables the authenticated user to unpin a List.

        :param user_id: The user ID who you are unpin a List on behalf of, with access token associated.
        :param list_id: The ID of the List that you would like the user id to unpin.
        :return: unpin list status.
        """
        resp = self._request(
            url=f"{self.BASE_URL_V2}/users/{user_id}/pinned_lists/{list_id}",
            verb="DELETE",
        )
        data = self._parse_response(resp=resp)
        return data

    def get_compliance_job(
        self,
        job_id: str,
        *,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Get a single compliance job with the specified ID.

        :param job_id: ID for the compliance job.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the job.
        """

        return self._get(
            url=f"{self.BASE_URL_V2}/compliance/jobs/{job_id}",
            params=None,
            cls=md.ComplianceJob,
            return_json=return_json,
        )

    def get_compliance_jobs(
        self,
        job_type: Optional[str],
        *,
        status: Optional[str] = None,
        return_json: bool = False,
    ) -> Union[dict, md.Response]:
        """
        Returns a list of recent compliance jobs.

        :param job_type: Type for the job, Accepted values are: tweets, users.
        :param status: Status for the job.
            Accepted values are: created, in_progress, failed, complete.
            Default is all.
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return:
            - data: data for the jobs.
        """

        args = {"type": job_type}
        if status:
            args["status"] = status

        return self._get(
            url=f"{self.BASE_URL_V2}/compliance/jobs",
            params=args,
            cls=md.ComplianceJob,
            multi=True,
            return_json=return_json,
        )

    def create_compliance_job(
        self,
        job_type: Optional[str],
        *,
        name: Optional[str] = None,
        resumable: Optional[bool] = None,
        return_json: bool = False,
    ) -> Union[dict, md.ComplianceJob]:
        """
        Creates a new compliance job for Tweet IDs or user IDs.
        You can run one batch job at a time.

        :param job_type: Type for the job, Accepted values are: tweets, users.
        :param name: A name for this job, useful to identify multiple jobs using a label you define.
        :param resumable: Specifies whether to enable the upload URL with support for resumable uploads
        :param return_json: Type for returned data. If you set True JSON data will be returned.
        :return: Compliance job information.
        """

        args = {"type": job_type}
        if name is not None:
            args["name"] = name
        if resumable is not None:
            args["resumable"] = resumable

        resp = self._request(
            url=f"{self.BASE_URL_V2}/compliance/jobs",
            verb="POST",
            json=args,
        )
        data = self._parse_response(resp=resp)
        if return_json:
            return data
        else:
            return md.ComplianceJob.new_from_json_dict(data["data"])



"""
    Poll object

    Refer: https://developer.twitter.com/en/docs/twitter-api/data-dictionary/object-model/poll
"""

from dataclasses import dataclass, field
from typing import List, Optional

from .base import BaseModel


@dataclass
class PollOption(BaseModel):
    position: Optional[int] = field(default=None, repr=False)
    label: Optional[str] = field(default=None)
    votes: Optional[int] = field(default=None, repr=False)


@dataclass
class Poll(BaseModel):
    """
    A class representing poll object.
    """

    id: Optional[str] = field(default=None)
    options: Optional[List[PollOption]] = field(default=None, repr=False)
    duration_minutes: Optional[int] = field(default=None, repr=False)
    end_datetime: Optional[str] = field(default=None, repr=False)
    voting_status: Optional[str] = field(default=None, repr=False)



"""
    Models for compliance
"""

from dataclasses import dataclass, field
from typing import Optional

from .base import BaseModel


@dataclass
class ComplianceJob(BaseModel):
    """
    A class representing the job for compliance.

    Refer: https://developer.twitter.com/en/docs/twitter-api/compliance/batch-compliance/api-reference/get-compliance-jobs-id
    """

    id: Optional[str] = field(default=None)
    created_at: Optional[str] = field(default=None, repr=False)
    type: Optional[str] = field(default=None)
    name: Optional[str] = field(default=None, repr=False)
    upload_url: Optional[str] = field(default=None, repr=False)
    upload_expires_at: Optional[str] = field(default=None, repr=False)
    download_url: Optional[str] = field(default=None, repr=False)
    download_expires_at: Optional[str] = field(default=None, repr=False)
    resumable: Optional[bool] = field(default=None, repr=False)
    status: Optional[str] = field(default=None)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytwitter', 'pytwitter.models', 'pytwitter.utils']

package_data = \
{'': ['*']}

install_requires = \
['Authlib>=0.15.4,<0.16.0',
 'dataclasses-json>=0.5.2,<0.6.0',
 'requests>=2.24,<3.0']

setup_kwargs = {
    'name': 'python-twitter-v2',
    'version': '0.7.2',
    'description': 'A simple Python wrapper for Twitter API v2 ✨ 🍰 ✨',
    'long_description': 'python-twitter\n\nA simple Python wrapper around for Twitter API v2 :sparkles: :cake: :sparkles:.\n\n.. image:: https://img.shields.io/endpoint?url=https%3A%2F%2Ftwbadges.glitch.me%2Fbadges%2Fv2\n   :target: https://developer.twitter.com/en/docs/twitter-api\n   :alt: v2\n\n.. image:: https://github.com/sns-sdks/python-twitter/workflows/Test/badge.svg\n    :target: https://github.com/sns-sdks/python-twitter/actions\n    :alt: Build Status\n\n.. image:: https://codecov.io/gh/sns-sdks/python-twitter/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/sns-sdks/python-twitter\n    :alt: Codecov\n\n.. image:: https://img.shields.io/pypi/v/python-twitter-v2.svg\n    :target: https://pypi.org/project/python-twitter-v2/\n    :alt: PyPI\n\n\n\n============\nIntroduction\n============\n\nTwitter has published new version `Twitter API V2 <https://twitter.com/TwitterDev/status/1293593516040269825>`_ for developer at Aug 13, 2020.\n\nThis library provides a service to easily use this new version Twitter API.\n\n=============\nDocumentation\n=============\n\nYou can get all api description and update at `Twitter API v2: Early Access <https://developer.twitter.com/en/docs/twitter-api/early-access>`_.\n\nLibrary docs site on `here <https://sns-sdks.github.io/python-twitter/>`_\n\n\n==========\nInstalling\n==========\n\nYou can install this library easily by `pypi`:\n\n.. code-block:: shell\n\n    $ pip install python-twitter-v2\n\nCode is hosted at `https://github.com/sns-sdks/python-twitter <https://github.com/sns-sdks/python-twitter>`_.\n\nCheckout latest development version with:\n\n.. code-block:: shell\n\n    $ git clone https://github.com/sns-sdks/python-twitter.git\n    $ cd python-twitter\n\nInstall dependencies with:\n\n.. code-block:: shell\n\n    $ make env\n\n\nRun tests with:\n\n.. code-block:: shell\n\n    $ make test\n\nRun tests with coverage:\n\n.. code-block:: shell\n\n    $ make cov-term\n    $ make cov-html\n\n=====\nUsing\n=====\n\nThe API is exposed via the ``pytwitter.Api`` class.\n\nNow cover features:\n\n- Tweets\n    - Tweet lookup\n    - Manage Tweets\n    - Timelines\n    - Search Tweets\n    - Tweet counts\n    - Filtered stream\n    - Volume streams\n    - Retweets\n    - Likes\n    - Hide replies\n\n- Users\n    - User lookup\n    - Follows\n    - Blocks\n    - Mutes\n\n- Spaces\n    - Spaces lookup\n    - Search Spaces\n\n- Compliance\n    - Batch compliance\n\n- Lists\n    - List lookup\n    - Manage lists\n    - List Tweets lookup\n    - List members\n    - List follows\n    - Pinned Lists\n\n-----------\nINSTANTIATE\n-----------\n\nYou can initialize with an bearer token:\n\n.. code-block:: python\n\n    >>> from pytwitter import Api\n    >>> api = Api(bearer_token="Your bearer token")\n\nWith user context token:\n\n.. code-block:: python\n\n    >>> api = Api(\n            consumer_key="consumer key",\n            consumer_secret="consumer secret",\n            access_token="access token",\n            access_secret="access secret"\n        )\n\nOr with authorize by user:\n\n.. code-block:: python\n\n    >>> api = Api(consumer_key="consumer key",consumer_secret="consumer secret",oauth_flow=True)\n    # get url for user to authorize\n    >>> api.get_authorize_url()\n    # copy the response url\n    >>> api.generate_access_token("https://localhost/?oauth_token=oauth_token&oauth_verifier=oauth_verifier")\n    {\'oauth_token\': \'oauth_token\',\n     \'oauth_token_secret\': \'oauth_token_secret\',\n     \'user_id\': \'123456\',\n     \'screen_name\': \'screen name\'}\n\nTwitter has `announcing OAuth 2.0 beta <https://twittercommunity.com/t/announcing-oauth-2-0-beta/159189>`_\n\nNow if you have app with ``OAuth2.0`` client ID. you can do authorize with ``OAuth2``.\n\n.. code-block:: python\n\n    >>> api = Api(client_id="You client ID", oauth_flow=True)\n    # get the url and code verifier for user to authorize\n    >>> url, code_verifier, _ = api.get_oauth2_authorize_url()\n    # copy the response url\n    >>> api.generate_oauth2_access_token("https://localhost/?state=state&code=code", code_verifier)\n    {\'token_type\': \'bearer\',\n     \'expires_in\': 7200,\n     \'access_token\': \'access_token\',\n     \'scope\': \'users.read tweet.read\',\n     \'expires_at\': 1631775928}\n\n\n------------\nUsers-lookup\n------------\n\nYou can get information about a user or group of users, specified by a user ID or a username.\n\nGet group of users:\n\n.. code-block:: python\n\n    # By ids\n    >>> api.get_users(ids=["783214", "2244994945"])\n    Response(data=[User(id=\'2244994945\', name=\'Twitter Dev\', username=\'TwitterDev\'), User(id=\'783214\', name=\'Twitter\', username=\'Twitter\')])\n\n    # By username\n    >>> api.get_users(usernames="Twitter,TwitterDev")\n    Response(data=[User(id=\'2244994945\', name=\'Twitter Dev\', username=\'TwitterDev\'), User(id=\'783214\', name=\'Twitter\', username=\'Twitter\')])\n\nGet single user:\n\n.. code-block:: python\n\n    # By id\n    >>> api.get_user(user_id="783214")\n    Response(data=User(id=\'783214\', name=\'Twitter\', username=\'Twitter\'))\n\n    # By username\n    >>> api.get_user(username="Twitter")\n    Response(data=User(id=\'783214\', name=\'Twitter\', username=\'Twitter\'))\n\nGet user following:\n\n.. code-block:: python\n\n    >>> api.get_following(user_id="2244994945", max_results=5)\n    Response(data=[User(id=\'459860328\', name=\'julie✨\', username=\'JulieMendoza206\'), User(id=\'273830767\', name=\'🄿🅄🅂🄷\', username=\'rahul_pushkarna\')...])\n\nGet user followers:\n\n.. code-block:: python\n\n    >>> api.get_followers(user_id="2244994945", max_results=5)\n    Response(data=[User(id=\'715131097332518912\', name=\'Daniel\', username=\'RGIDaniel\'), User(id=\'1176323137757048832\', name=\'Joyce Wang\', username=\'joycew67\')...])\n\n\nYou can follow or unfollow user if you have OAuth 1.0a User context.\n\nfollow user:\n\n.. code-block:: python\n\n    >>> api.follow_user(user_id="123456", target_user_id="654321")\n    {\'data\': {\'following\': True, \'pending_follow\': False}}\n\n\nunfollow user:\n\n.. code-block:: python\n\n    >>> api.unfollow_user(user_id="123456", target_user_id="654321")\n    {\'data\': {\'following\': False}}\n\n-------------\nTweets-lookup\n-------------\n\nYou can get information about a tweet or group of tweets by tweet id(s).\n\nGet single tweet:\n\n.. code-block:: python\n\n    >>> api.get_tweet("1354143047324299264", expansions=["attachments.media_keys"], media_fields=["type","duration_ms"])\n    Response(data=Tweet(id=1354143047324299264, text=Academics are one of the biggest groups using...))\n\nGet group of tweets:\n\n.. code-block:: python\n\n    >>> api.get_tweets(["1261326399320715264","1278347468690915330"],expansions="author_id",tweet_fields=["created_at"], user_fields=["username","verified"])\n    Response(data=[Tweet(id=1261326399320715264, text=Tune in to the @MongoDB @Twitch stream...), Tweet(id=1278347468690915330, text=Good news and bad news: 2020 is half over)])\n\n-------------\nStreaming API\n-------------\n\nFor Streaming, this provide `StreamApi` independent. Same as main `Api`, You need initial it first.\n\n.. code-block:: python\n\n    >>> from pytwitter import StreamApi\n    >>> stream_api = StreamApi(bearer_token="bearer token")\n    # or use consumer key and secret\n    >>> stream_api = StreamApi(consumer_key="consumer key", consumer_secret="consumer secret")\n\n\nFor Sample Stream tweets, You can use `sample_stream` function to build a connection.\n\n.. code-block:: python\n\n    >>> stream_api.sample_stream()\n\nFor Search Stream, You can point your rules.\n\nGet your current rules.\n\n.. code-block:: python\n\n    >>> stream_api.get_rules()\n    Response(data=[StreamRule(id=\'1369580714056843266\', value=\'twitter api \')])\n\nDelete You rules.\n\n.. code-block:: python\n\n    >>> stream_api.manage_rules(rules={"delete": {"ids": ["1369580714056843266"]}})\n    Response(data=[])\n\nAdd new rules. If you set `dry_run` to True, will only validate rules, and not create them.\n\n.. code-block:: python\n\n    >>> np = {\n            "add": [\n                {"value": "cat has:media", "tag": "cats with media"},\n                {"value": "cat has:media -grumpy", "tag": "happy cats with media"}\n            ]\n         }\n    >>> stream_api.manage_rules(rules=np, dry_run=True)\n    Response(data=[StreamRule(id=\'1370406958721732610\', value=\'cat has:media -grumpy\'), StreamRule(id=\'1370406958721732609\', value=\'cat has:media\')])\n\nThen you can use `search_stream` to get tweets match your rules.\n\n.. code-block:: python\n\n    >>> stream_api.search_stream()\n\n\nYou can go to the `Example folder <examples>`_ for streaming examples.\n\n====\nTODO\n====\n\n- More Api waiting twitter\n\n',
    'author': 'ikaroskun',
    'author_email': 'merle.liukun@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sns-sdks/python-twitter',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from typing import (
    Dict,
    Type,
    TypeVar,
    Optional,
)

from dataclasses_json import (
    DataClassJsonMixin,
)

A = TypeVar("A", bound=DataClassJsonMixin)


@dataclass
class BaseModel(DataClassJsonMixin):
    @classmethod
    def new_from_json_dict(
        cls: Type[A], data: Optional[Dict], *, infer_missing=False
    ) -> Optional[A]:
        """
        Convert json dict to data class
        :param data: A json dict which will convert model class.
        :param infer_missing: if set True, will let missing field (not have default vale) to None
        :return: The data class
        """
        if not data:
            return None
        c = cls.from_dict(data, infer_missing=infer_missing)
        # save origin data
        cls._json = data
        return c
